import os
import subprocess
from io import StringIO
from rkd.inputoutput import BufferedSystemIO
from rkd.inputoutput import IO
from rkd.contract import ExecutionContext
from rkd.syntax import TaskDeclaration
from harbor.test import BaseHarborTestClass
from harbor.test import TestTask


class TestHooksFeature(BaseHarborTestClass):
    def _prepare_test_data(self):
        with open(self.get_test_env_subdirectory('hooks.d/pre-upgrade') + '/whoami.sh', 'w') as fw:
            fw.write('''
                #!/bin/bash
                echo " >> This is a whoami.sh hook, test: $(whoami)"
            ''')

        with open(self.get_test_env_subdirectory('hooks.d/post-upgrade') + '/history.sh', 'w') as fw:
            fw.write('''
                #!/bin/bash
                echo "25 June 1978 the rainbow flag was first flown at the San Francisco Gay Freedom Day Parade. The flag was designed by artist Gilbert Baker, and hand-dyed and stitched by 30 volunteers."
            ''')

        subprocess.check_call(['chmod', '+x', self.get_test_env_subdirectory('hooks.d/pre-upgrade') + '/whoami.sh'])
        subprocess.check_call(['chmod', '+x', self.get_test_env_subdirectory('hooks.d/post-upgrade') + '/history.sh'])

    def test_functional_hooks_are_executed_when_exists_and_files_with_extension_only_are_skipped(self):
        """Given we have an example hooks in pre-upgrade/whoami.sh and in post-upgrade/history.sh
        And we try to run those hooks using hooks_executed()
        Then we will see output produced by those scripts
        And .dotfiles will be ignored
        """

        self._prepare_test_data()

        buffer = StringIO()
        hooks_capturing_io = IO()

        task = TestTask()
        task._io = BufferedSystemIO()
        ctx = ExecutionContext(
            TaskDeclaration(task),
            args={},
            env={}
        )

        with hooks_capturing_io.capture_descriptors(stream=buffer, enable_standard_out=True):
            with task.hooks_executed(ctx, 'upgrade'):
                pass

        self.assertIn('>> This is a whoami.sh hook, test:', buffer.getvalue(),
                      msg='Expected pre-upgrade hook to be ran')

        self.assertIn('25 June 1978 the rainbow flag was first flown', buffer.getvalue(),
                      msg='Expected post-upgrade hook to be ran')

        self.assertIn('pre-upgrade/whoami.sh', task._io.get_value())
        self.assertNotIn('.gitkeep', task._io.get_value())

    def test_functional_execute_hooks_executes_post_upgrade_hooks(self):
        """Assert that post-upgrade/history.sh is executed"""

        self._prepare_test_data()

        buffer = StringIO()
        hooks_capturing_io = IO()

        task = TestTask()
        task._io = BufferedSystemIO()
        ctx = ExecutionContext(
            TaskDeclaration(task),
            args={},
            env={}
        )

        with hooks_capturing_io.capture_descriptors(stream=buffer, enable_standard_out=True):
            task.execute_hooks(ctx, 'post-upgrade')

        self.assertIn('25 June 1978 the rainbow flag was first flown', buffer.getvalue(),
                      msg='Expected post-upgrade hook to be ran')

    def test_non_existing_dir_is_skipped(self):
        """Assert that non-existing directory does not cause exception, but will be skipped"""

        task = TestTask()
        task._io = BufferedSystemIO()
        task._io.set_log_level('debug')
        ctx = ExecutionContext(
            TaskDeclaration(task),
            args={},
            env={}
        )

        task.execute_hooks(ctx, 'non-existing-directory')
        self.assertIn('Hooks dir "./hooks.d//non-existing-directory/" not present, skipping', task._io.get_value())




import os
import sys
import unittest
import subprocess
import time
import yaml
from io import StringIO
from typing import Dict
from copy import deepcopy
from argparse import ArgumentParser
from rkd.contract import ExecutionContext
from rkd.context import ApplicationContext
from rkd.executor import OneByOneTaskExecutor
from rkd.inputoutput import IO
from rkd.inputoutput import BufferedSystemIO
from rkd.syntax import TaskDeclaration
from .tasks.base import HarborBaseTask
from .service import ServiceDeclaration
from .driver import ComposeDriver
from .cached_loader import CachedLoader
from dotenv import dotenv_values

HARBOR_MODULE_PATH = os.path.dirname(os.path.realpath(__file__))
ENV_SIMPLE_PATH = os.path.dirname(os.path.realpath(__file__)) + '/../../test/testdata/env_simple'
CURRENT_TEST_ENV_PATH = os.path.dirname(os.path.realpath(__file__)) + '/../../test/testdata/current_test_env'
TEST_PROJECT_NAME = 'env_simple'


class TestTask(HarborBaseTask):
    is_dev_env = True

    def get_name(self) -> str:
        return ':test'

    def get_group_name(self) -> str:
        return ''

    def execute(self, context: ExecutionContext) -> bool:
        return True

    def run(self, context: ExecutionContext) -> bool:
        return True

    def configure_argparse(self, parser: ArgumentParser):
        pass


def create_mocked_task(io: IO) -> TestTask:
    task = TestTask()
    ctx = ApplicationContext([], [], '')
    ctx.io = io

    task.internal_inject_dependencies(
        io=io,
        ctx=ctx,
        executor=OneByOneTaskExecutor(ctx=ctx)
    )

    return task


class BaseHarborTestClass(unittest.TestCase):
    _stderr_bckp = None
    _stdout_bckp = None

    def setUp(self) -> None:
        print('')
        print('==================================================================================================' +
              '=====================================')
        print('Test name: ' + self.__class__.__name__ + ' :: ' + self._testMethodName)
        print('----------')
        print('')

        CachedLoader.clear()   # avoid keeping the state between tests

        os.chdir(HARBOR_MODULE_PATH)
        self.recreate_structure()
        self.setup_environment()
        self.remove_all_containers()
        self._stderr_bckp = sys.stderr
        self._stdout_bckp = sys.stdout

    def tearDown(self) -> None:
        if sys.stderr != self._stderr_bckp or sys.stdout != self._stdout_bckp:
            print('!!! Test ' + self.id() + ' is not cleaning up stdout/stderr')

        self._restore_streams()

    @classmethod
    def mock_compose(cls, content: dict):
        content['version'] = '3.4'

        with open(CURRENT_TEST_ENV_PATH + '/apps/conf/mocked.yaml', 'wb') as f:
            f.write(yaml.dump(content).encode('utf-8'))

    def _restore_streams(self):
        sys.stderr = self._stderr_bckp
        sys.stdout = self._stdout_bckp

    @classmethod
    def recreate_structure(cls):
        """Within each class recreate the project structure, as it could be changed by tests itself"""

        subprocess.check_call(['rm', '-rf', CURRENT_TEST_ENV_PATH])
        subprocess.check_call(['cp', '-pr', ENV_SIMPLE_PATH, CURRENT_TEST_ENV_PATH])

        # copy from base structure - as we test eg. things like default configurations, NGINX template
        for directory in ['containers', 'data', 'hooks.d', 'apps/www-data']:
            subprocess.check_call('rm -rf %s/%s' % (ENV_SIMPLE_PATH, directory), shell=True)
            subprocess.check_call('cp -pr %s/project/%s %s/%s' % (
                HARBOR_MODULE_PATH, directory, CURRENT_TEST_ENV_PATH, directory
            ), shell=True)

        cls.mock_compose({'services': {}})

    @classmethod
    def get_test_env_subdirectory(cls, subdir_name: str):
        directory = CURRENT_TEST_ENV_PATH + '/' + subdir_name

        if not os.path.isdir(directory):
            subprocess.check_call(['mkdir', '-p', directory])

        return os.path.realpath(directory)

    @classmethod
    def remove_all_containers(cls):
        try:
            subprocess.check_output("docker rm -f -v $(docker ps -a --format '{{ .Names }}' | grep " + TEST_PROJECT_NAME + ")",
                                    shell=True, stderr=subprocess.STDOUT)

        except subprocess.CalledProcessError as e:
            # no containers found - it's OK
            if "requires at least 1 argument" in str(e.output):
                return

            raise e

    @classmethod
    def setup_environment(cls):
        os.environ.update(dotenv_values(CURRENT_TEST_ENV_PATH + '/.env'))
        os.environ['APPS_PATH'] = CURRENT_TEST_ENV_PATH + '/apps'
        os.environ['RKD_PATH'] = cls.get_test_env_subdirectory('') + ':' + HARBOR_MODULE_PATH + '/internal'

        os.chdir(CURRENT_TEST_ENV_PATH)

    def _get_prepared_compose_driver(self, args: dict = {}, env: dict = {}) -> ComposeDriver:
        merged_env = deepcopy(os.environ)
        merged_env.update(env)

        task = create_mocked_task(BufferedSystemIO())
        declaration = TaskDeclaration(task)
        ctx = ExecutionContext(declaration, args=args, env=merged_env)

        return ComposeDriver(task, ctx, TEST_PROJECT_NAME)

    def execute_task(self, task: HarborBaseTask, args: dict = {}, env: dict = {}) -> str:
        ctx = ApplicationContext([], [], '')
        ctx.io = BufferedSystemIO()

        task.internal_inject_dependencies(
            io=ctx.io,
            ctx=ctx,
            executor=OneByOneTaskExecutor(ctx=ctx)
        )

        merged_env = deepcopy(os.environ)
        merged_env.update(env)

        r_io = IO()
        str_io = StringIO()

        with r_io.capture_descriptors(enable_standard_out=True, stream=str_io):
            result = task.execute(ExecutionContext(
                TaskDeclaration(task),
                args=args,
                env=merged_env
            ))

        return ctx.io.get_value() + "\n" + str_io.getvalue() + "\nTASK_EXIT_RESULT=" + str(result)

    @staticmethod
    def prepare_service_discovery(driver: ComposeDriver):
        driver.up(ServiceDeclaration('gateway', {}), capture=True, force_recreate=True)
        driver.up(ServiceDeclaration('gateway_proxy_gen', {}), capture=True, force_recreate=True)
        driver.up(ServiceDeclaration('website', {}), capture=True)

    def prepare_example_service(self, name: str, uses_service_discovery: bool = False) -> ComposeDriver:
        drv = self._get_prepared_compose_driver()

        # prepare
        drv.rm(ServiceDeclaration(name, {}))
        drv.up(ServiceDeclaration(name, {}))

        if uses_service_discovery:
            # give service discovery some time
            # @todo: This can be improved possibly
            time.sleep(5)

        return drv

    def get_containers_state(self, driver: ComposeDriver) -> Dict[str, bool]:
        running_rows = driver.scope.sh('docker ps -a --format "{{ .Names }}|{{ .Status }}"', capture=True).split("\n")
        containers = {}

        for container_row in running_rows:
            try:
                name, status = container_row.split('|')
            except ValueError:
                continue

            if name.startswith(driver.project_name + '_'):
                containers[name] = 'Up' in status

        return containers

    def get_locally_pulled_docker_images(self) -> list:
        images = subprocess.check_output(['docker', 'images', '--format', '{{ .Repository }}:{{ .Tag }}'])\
            .decode('utf-8')\
            .split("\n")

        return images

    def exec_in_container(self, container_name: str, cmd: list) -> str:
        return subprocess.check_output(
            ['docker', 'exec', '-i', container_name] + cmd,
            stderr=subprocess.STDOUT
        ).decode('utf-8')

    def fetch_page_content(self, host: str):
        return self.exec_in_container(TEST_PROJECT_NAME + '_gateway_1', ['curl', '-s', '-vv', '--header',
                                                                         'Host: %s' % host, 'http://127.0.0.1'])

    def prepare_valid_deployment_yml(self):
        """Internal: Prepares a valid deployment.yml and already synchronized files structure, downloaded role"""

        with open(self.get_test_env_subdirectory('') + '/deployment.yml', 'w') as f:
            f.write('''deploy_user: vagrant
deploy_group: vagrant
remote_dir: /project
git_url: https://github.com/riotkit-org/empty
git_secret_url: https://github.com/riotkit-org/empty
configure_sudoers: no

nodes:
    production:
        # example configuration for testing environment based on Vagrant
        # to run the environment type: harbor :deployment:vagrant -c "up --provision"
        - host: 127.0.0.1
          port: 2222
          user: docker
          password: docker
''')

    def assertContainerIsNotRunning(self, service_name: str, driver: ComposeDriver):
        container_name_without_instance_num = driver.project_name + '_' + service_name + '_'

        for name, state in self.get_containers_state(driver).items():
            if name.startswith(container_name_without_instance_num) and state is True:
                self.fail('"%s" is running, but should not' % name)

    def assertLocalRegistryHasImage(self, image_name):
        self.assertIn(image_name, self.get_locally_pulled_docker_images(),
                      msg='Expected that "docker images" will contain image "%s"' % image_name)

    def assertLocalRegistryHasNoPulledImage(self, image_name):
        self.assertNotIn(image_name, self.get_locally_pulled_docker_images(),
                         msg='Expected that "docker images" will not contain image "%s"' % image_name)



import os
from harbor.test import BaseHarborTestClass
from harbor.tasks.deployment import UpdateFilesTask


class UpdateFilesTaskTest(BaseHarborTestClass):
    def test_functional_validates_deployment_yml_file_not_created(self):
        """Test that deployment.yml/deployment.yaml needs to be created first and the message for the end user is clear
        """

        out = self.execute_task(UpdateFilesTask(), args={'--ask-vault-pass': False, '--vault-passwords': ''}, env={})
        self.assertIn('Deployment not configured - missing deployment.yml or deployment.yaml file', out)

    def test_functional_structure_is_copied(self):
        """Verify that the structure is copied, and the files are rendered using JINJA2"""

        # prepare configuration
        self.prepare_valid_deployment_yml()
        task = UpdateFilesTask()

        # mock external dependencies: github.com and ansible-galaxy
        task.download_roles = lambda *args, **kwargs: None

        with self.subTest('Verify overall task exit code'):
            out = self.execute_task(task, args={'--ask-vault-pass': False, '--vault-passwords': ''}, env={})
            self.assertIn('TASK_EXIT_RESULT=True', out)

        with self.subTest('Verify copied files'):
            # check a few files, including files that are rendered from JINJA2 like inventory and playbook
            for filename in ['ansible.cfg', 'Vagrantfile', 'harbor.inventory.cfg', 'harbor.playbook.yml']:
                self.assertTrue(os.path.isfile(self.get_test_env_subdirectory('.rkd/deployment') + '/' + filename))

        with self.subTest('Check if JINJA2 templates are rendered'):
            with open(self.get_test_env_subdirectory('.rkd/deployment') + '/harbor.inventory.cfg', 'r') as f:
                # vagrant is a default value from file created by self.prepare_valid_deployment_yml()
                self.assertIn('ansible_ssh_user=docker', f.read())




import requests
from time import time
from io import StringIO
from rkd.inputoutput import IO
from harbor.test import BaseHarborTestClass
from harbor.service import ServiceDeclaration
from harbor.exception import ServiceNotCreatedException
from harbor.exception import ServiceNotReadyException


class ComposeDriverTest(BaseHarborTestClass):
    def test_get_compose_args_sets_basic_arguments_and_sees_yaml_configuration_files(self):
        drv = self._get_prepared_compose_driver()
        args = drv.get_compose_args()

        self.assertIn('--project-directory=', args, msg='Project directory should be set')
        self.assertIn('-p env_simple', args, msg='Project name should be set')
        self.assertIn('-f docker-compose.yml', args, msg='Basic docker-compose.yml should be considered')
        self.assertIn('infrastructure.ssl.yml', args)
        self.assertIn('infrastructure.service-discovery.yml', args)
        self.assertIn('infrastructure.health.yml', args)

    def test_ps(self):
        """Simply check if docker-compose ps can be executed, if the standard switches are correct"""

        io_str = StringIO()
        io = IO()

        with io.capture_descriptors(stream=io_str, enable_standard_out=False):
            drv = self._get_prepared_compose_driver()
            drv.ps([])

        self.assertIn('Name', io_str.getvalue())
        self.assertIn('Ports', io_str.getvalue())

    def test_up(self):
        """Simply verify that the docker-compose up command is properly generated"""

        resulting_args = []

        # mock
        drv = self._get_prepared_compose_driver()
        drv.scope.sh = lambda *args, **kwargs: (resulting_args.append(args))

        # action
        drv.up(ServiceDeclaration('healthcheck', {}))

        args_as_str = ' '.join(resulting_args[0])

        self.assertIn('--scale healthcheck=1 healthcheck', args_as_str)
        self.assertIn('infrastructure.health.yml', args_as_str)
        self.assertIn('up -d', args_as_str)

    def test_up_invalid_parameters_for_recreation(self):
        """Unit test: Checks if exception is raised when logic error happens"""

        drv = self._get_prepared_compose_driver()
        drv.scope.sh = lambda *args, **kwargs: ''

        # action
        self.assertRaises(Exception, lambda: drv.up(ServiceDeclaration('healthcheck', {}),
                                                    norecreate=True, force_recreate=True))

    def test_up_tolerates_extra_arguments(self):
        """Unit tests: Checks if extra arguments are allowed to be passed"""

        resulting_args = []
        drv = self._get_prepared_compose_driver()
        drv.scope.sh = lambda *args, **kwargs: (resulting_args.append(args))

        # action
        drv.up(ServiceDeclaration('healthcheck', {}), extra_args='--remove-orphans')
        args_as_str = ' '.join(resulting_args[0])

        self.assertIn('--remove-orphans', args_as_str)

    def test_functional_up_and_down(self):
        """Functional test to start, stop, restart, remove container
           Checks basic compatibility with docker-compose interface
        """

        drv = self._get_prepared_compose_driver()

        # 1) Start
        drv.up(ServiceDeclaration('gateway', {}))
        self.assertTrue(self.get_containers_state(drv)['env_simple_gateway_1'])

        # 2) Restart
        drv.restart('gateway')
        self.assertTrue(self.get_containers_state(drv)['env_simple_gateway_1'])

        # 3) Stop
        drv.stop('gateway')
        self.assertFalse(self.get_containers_state(drv)['env_simple_gateway_1'])

        # 4) Remove
        drv.rm(ServiceDeclaration('gateway', {}))
        self.assertNotIn('env_simple_gateway_1', self.get_containers_state(drv))

    def test_get_last_container_name_for_service(self):
        drv = self._get_prepared_compose_driver()

        # make sure it is clean up
        drv.rm(ServiceDeclaration('gateway', {}), capture=True)  # there may be more instances, so remove them
        drv.up(ServiceDeclaration('gateway', {}), capture=True)  # bring one instance

        self.assertEqual('env_simple_gateway_1', drv.get_last_container_name_for_service('gateway'))

    def test_find_container_name_case_service_not_created_raises_exception(self):
        drv = self._get_prepared_compose_driver()

        self.assertRaises(
            ServiceNotCreatedException,
            lambda: drv.find_container_name(ServiceDeclaration('not_existing', {}))
        )

    def test_find_container_name_finds_first_instance_when_no_instance_specified(self):
        drv = self.prepare_example_service('gateway')

        instance_name = drv.find_container_name(ServiceDeclaration('gateway', {}))
        self.assertEqual('env_simple_gateway_1', instance_name)

    def test_find_container_name_finds_first_instance_by_specyfing_instance_num(self):
        drv = self.prepare_example_service('gateway')

        self.assertEqual(
            'env_simple_gateway_1',
            drv.find_container_name(ServiceDeclaration('gateway', {}), instance_num=1)
        )

    def test_find_container_name_does_not_find_instance_when_there_is_not_enough_replicas(self):
        drv = self.prepare_example_service('gateway')

        # for single works
        self.assertEqual('env_simple_gateway_1', drv.find_container_name(ServiceDeclaration('gateway', {}), instance_num=1))

        self.assertRaises(
            ServiceNotCreatedException,
            lambda: drv.find_container_name(ServiceDeclaration('gateway', {}), instance_num=50)
        )

    def test_wait_for_log_message_finds_searched_phrase(self):
        drv = self.prepare_example_service('gateway')

        requests.get('http://localhost:8000')
        self.assertTrue(
            drv.wait_for_log_message('GET / HTTP/1.1', ServiceDeclaration('gateway', {}), timeout=5)
        )

    def test_wait_for_log_hits_timeout(self):
        drv = self.prepare_example_service('gateway')

        start_time = time()

        self.assertRaises(
            ServiceNotReadyException,
            lambda: drv.wait_for_log_message(
                'Some message that should be there',
                ServiceDeclaration('gateway', {}),
                timeout=2
            )
        )

        self.assertGreaterEqual(time() - start_time, 2, msg='Expected at least two seconds waiting')

    def test_scale_one_up_then_scale_down_to_desired_state(self):
        """Functional test: Scale one up, then down to declared state

        Covers:
            scale_one_up()
            scale_to_desired_state()
        """

        drv = self.prepare_example_service('website')
        drv.scale_one_up(ServiceDeclaration('website', {}))

        running_containers = self.get_containers_state(drv)
        self.assertIn('env_simple_website_1', running_containers)
        self.assertIn('env_simple_website_2', running_containers)

    def test_find_all_container_names_for_service_finds_containers(self):
        drv = self.prepare_example_service('website')

        self.assertEqual(['env_simple_website_1'],
                         drv.find_all_container_names_for_service(ServiceDeclaration('website', {})))

        drv.scale_one_up(ServiceDeclaration('website', {}))
        containers = drv.find_all_container_names_for_service(ServiceDeclaration('website', {}))

        self.assertIn('env_simple_website_1', containers)
        self.assertIn('env_simple_website_2', containers)

    def test_find_all_container_names_for_service_raises_exception_on_invalid_service_name(self):
        drv = self._get_prepared_compose_driver()

        self.assertRaises(
            ServiceNotCreatedException,
            lambda: drv.find_all_container_names_for_service(ServiceDeclaration('not_existing', {}))
        )

    def test_get_created_containers_is_returning_containers_in_proper_order(self):
        """Test that get_created_containers() is preserving order"""

        drv = self._get_prepared_compose_driver()

        # create three instances of a service
        drv.up(ServiceDeclaration('website', {}))
        drv.scale_one_up(ServiceDeclaration('website', {}))

        containers = drv.get_created_containers(only_running=False)
        self.assertEqual([1, 2], list(containers['website'].keys()))

    def test_get_created_containers_is_showing_only_running_containers(self):
        """Test that 'website' service will be created, but will have no active containers"""

        drv = self._get_prepared_compose_driver()

        # create three instances of a service
        drv.up(ServiceDeclaration('website', {}))
        drv.stop('website')

        containers = drv.get_created_containers(only_running=True)
        self.assertEqual([], list(containers['website']))



from subprocess import CalledProcessError
from harbor.test import BaseHarborTestClass
from harbor.tasks.repositories import FetchRepositoryTask
from harbor.tasks.repositories import ListRepositoriesTask
from harbor.tasks.repositories import FetchAllRepositories


class TestRepositories(BaseHarborTestClass):
    def test_functional_repository_is_cloned_then_pulled(self):
        """Scenario: Clone repository, then try to clone again - it should end with a pull of existing repository

        Additionally checks that pre_update() and post_update() hooks are executed.
        """

        with self.subTest('Test clone'):
            out = self.execute_task(FetchRepositoryTask(), args={'name': 'example'})
            self.assertIn('Cloning into', out)
            self.assertIn('TASK_EXIT_RESULT=True', out)

            # verify that hooks were executed at all
            self.assertIn('Starting to clone/pull example repository :)', out)
            self.assertIn("I'm a post_update hook", out)

        with self.subTest('Test checkout'):
            out = self.execute_task(FetchRepositoryTask(), args={'name': 'example'})
            self.assertIn('Already up to date', out)
            self.assertIn('TASK_EXIT_RESULT=True', out)

            # verify that hooks were executed at all
            self.assertIn('Starting to clone/pull example repository :)', out)
            self.assertIn("I'm a post_update hook", out)

    def test_will_not_perform_action_for_unknown_application_and_return_meaningful_error(self):
        """Check that application name is validated, and a tip is displayed with an expected path to the configuration
        file"""

        out = self.execute_task(FetchRepositoryTask(), args={'name': 'not_working'})

        self.assertIn('Cannot pull a repository: Unknown application', out)
        self.assertIn('repos-enabled/not_working.sh" file not found', out)

    def test_repository_listing_shows_example_application_repository(self):
        """Assert that example data appears"""

        out = self.execute_task(ListRepositoriesTask(), args={})

        self.assertIn('Configured GIT repository', out)
        self.assertIn('example', out)
        self.assertIn('second', out)

    def test_fetch_all_repositories_is_skipping_failures_and_keeping_going_on_but_producing_a_failure_at_the_end(self):
        """Check that FetchAllRepositories() is fetching all repositories one-by-one, ignoring a failure,
        but marking the whole operation as failed if at least one repository failed to update"""

        collected_args = []
        task = FetchAllRepositories()

        def mocked_rkd(*args, **kwargs):
            if len(collected_args) == 0:
                collected_args.append(list(args)[0])
                raise CalledProcessError(returncode=128, cmd='Hello, this method was mocked ;)')

            collected_args.append(list(args)[0])

        task.rkd = mocked_rkd
        out = self.execute_task(task, args={})

        # check logging
        self.assertIn('Updating "example"', out)
        self.assertIn('Updating "second"', out)
        self.assertIn('Failed updating "example"', out)
        self.assertIn('TASK_EXIT_RESULT=False', out)

        self.assertEqual(
            [[':harbor:git:apps:update', '--name=example'], [':harbor:git:apps:update', '--name=second']],
            collected_args
        )



import os
import subprocess
from uuid import uuid4
from abc import ABC
from jinja2 import Environment
from jinja2 import FileSystemLoader
from jinja2 import StrictUndefined
from jinja2.exceptions import UndefinedError
from typing import Dict
from typing import Tuple
from argparse import ArgumentParser
from rkd.contract import ExecutionContext
from rkd.yaml_parser import YamlFileLoader
from rkd.exception import MissingInputException
from .base import HarborBaseTask
from ..formatting import development_formatting
from ..exception import MissingDeploymentConfigurationError

HARBOR_ROOT = os.path.dirname(os.path.realpath(__file__)) + '/../deployment/files'


class BaseDeploymentTask(HarborBaseTask, ABC):
    ansible_dir: str = '.rkd/deployment'
    _config: dict
    vault_args: list = []

    def get_config(self) -> dict:
        """Loads and parses deployment.yml file. Supports Ansible Vault encryption"""

        deployment_filenames = ['deployment.yml', 'deployment.yaml']

        try:
            self._config
        except AttributeError:

            # try multiple files
            for filename in deployment_filenames:
                if os.path.isfile(filename):

                    #
                    # Check file contents before
                    #
                    with open(filename, 'rb') as f:
                        content = f.read().decode('utf-8')
                        #
                        # When file is encrypted, then decrypt it
                        #
                        if content.startswith('$ANSIBLE_VAULT;'):
                            tmp_vault_filename = '.tmp-' + str(uuid4())
                            tmp_vault_path = './.rkd/' + tmp_vault_filename

                            self.sh('cp %s %s' % (filename, tmp_vault_path))
                            self.rkd([':harbor:vault:encrypt', '-d', tmp_vault_path] + self.vault_args)

                            try:
                                self._config = YamlFileLoader(self._ctx.directories).load_from_file(
                                    tmp_vault_filename,
                                    'org.riotkit.harbor/deployment/v1'
                                )
                            finally:
                                self.sh('rm -f %s' % tmp_vault_path)

                            return self._config

                    self._config = YamlFileLoader(self._ctx.directories).load_from_file(
                        filename,
                        'org.riotkit.harbor/deployment/v1'
                    )

                    return self._config

            raise MissingDeploymentConfigurationError()

        return self._config

    def _verify_synced_version(self, abs_ansible_dir: str):
        """Verifies last synchronization - displays warning if Harbor version was changed after last
        files synchronization"""

        if not os.path.isfile(abs_ansible_dir + '/.synced'):
            return

        with open(abs_ansible_dir + '/.synced', 'rb') as f:
            synced_version = f.read().decode('utf-8').strip()
            actual_version = self.get_harbor_version()

            if synced_version != actual_version:
                self.io().warn('Ansible deployment in .rkd/deployment is not up-to-date. We recommend to update' +
                               ' from %s to %s' % (synced_version, actual_version))

    def _write_synced_version(self, abs_ansible_dir: str):
        """Writes information about, in which Harbor version the files were synced last time"""

        with open(abs_ansible_dir + '/.synced', 'wb') as f:
            f.write(self.get_harbor_version().encode('utf-8'))

    def role_is_installed_and_configured(self) -> bool:
        return os.path.isfile(self.ansible_dir + '/.synced')

    def install_and_configure_role(self, force_update: bool = False) -> bool:
        """Install an Ansible role from galaxy, and configure playbook, inventory, all the needed things"""

        abs_ansible_dir = os.path.realpath(self.ansible_dir)
        should_update = force_update or not os.path.isfile(abs_ansible_dir + '/.synced')

        self.io().info('Checking role installation...')
        self._silent_mkdir(abs_ansible_dir)
        self._verify_synced_version(abs_ansible_dir)

        if not self._synchronize_structure_from_template(abs_ansible_dir, only_jinja_templates=True):
            self.io().error_msg('Cannot synchronize templates')
            return False

        if should_update:
            self.io().info('Role will be updated')

            if not self._synchronize_structure_from_template(abs_ansible_dir):
                self.io().error_msg('Cannot synchronize structure')
                return False

            self.io().debug('Downloading fresh role...')
            self.download_roles()

            self._write_synced_version(abs_ansible_dir)

        return True

    def download_roles(self):
        self.sh(' '.join([
            'ansible-galaxy',
            'install', '-r', self.ansible_dir + '/requirements.yml',
            '-p', self.ansible_dir + '/roles/',
            '--force'
        ]), capture=False)

    def _synchronize_structure_from_template(self, abs_ansible_dir: str, only_jinja_templates: bool = False) -> bool:
        """Synchronizes template structure into .rkd/deployment"""

        self.io().debug(
            'Synchronizing structure from template (only_jinja_templates=' + str(only_jinja_templates) + ')')

        # synchronize directory structure
        for root, subdirs, files in os.walk(HARBOR_ROOT):
            relative_root = root[len(HARBOR_ROOT) + 1:]

            self._silent_mkdir(abs_ansible_dir + '/' + relative_root)

            for file in files:
                if only_jinja_templates and not file.endswith('.j2'):
                    continue

                abs_src_file_path = root + '/' + file
                abs_dest_file_path = abs_ansible_dir + '/' + relative_root + '/' + file

                if not self._copy_file(abs_src_file_path, abs_dest_file_path):
                    self.io().error('Cannot process file %s' % abs_dest_file_path)
                    return False

        return True

    def _copy_file(self, abs_src_file_path: str, abs_dest_file_path: str):
        """Copies a file from template directory - supports jinja2 files rendering on-the-fly"""

        if abs_dest_file_path.endswith('.j2'):
            abs_dest_file_path = abs_dest_file_path[:-3]

            with open(abs_src_file_path, 'rb') as f:
                tpl = Environment(loader=FileSystemLoader(['./', './rkd/deployment']), undefined=StrictUndefined)\
                        .from_string(f.read().decode('utf-8'))

            try:
                variables = self._prepare_variables()

                with open(abs_dest_file_path, 'wb') as f:
                    f.write(tpl.render(**variables).encode('utf-8'))

            except UndefinedError as e:
                self.io().error(str(e) + " - required in " + abs_src_file_path + ", please define it in deployment.yml")
                return False

            return True

        subprocess.check_call(['cp', '-p', abs_src_file_path, abs_dest_file_path])
        self.io().debug('Created ' + abs_dest_file_path)
        return True

    def _prepare_variables(self):
        """Glues together variables from environment and from deployment.yaml for exposing in JINJA2 templates"""

        variables = {}
        variables.update(os.environ)
        variables.update(self.get_config())

        if 'git_url' not in variables:
            variables['git_url'] = subprocess\
                .check_output(['git', 'config', '--get', 'remote.origin.url']).decode('utf-8')\
                .replace('\n', '')\
                .strip()

        if 'git_secret_url' not in variables:
            variables['git_secret_url'] = variables['git_url'].replace('\n', '')

        return variables

    def _preserve_vault_parameters_for_usage_in_inner_tasks(self, ctx: ExecutionContext):
        """Preserve original parameters related to Vault, so those parameters can be propagated to inner tasks"""

        try:
            vault_passwords = ctx.get_arg_or_env('--vault-passwords')
        except MissingInputException:
            vault_passwords = ''

        # keep the vault arguments for decryption of deployment.yml
        self.vault_args = ['--vault-passwords=' + vault_passwords]
        if ctx.get_arg('--ask-vault-pass'):
            self.vault_args += '--ask-vault-pass'

    def _get_vault_opts(self, ctx: ExecutionContext, chdir: str = '') -> str:
        """Creates options to pass in Ansible Vault commandline"""

        try:
            vault_passwords = ctx.get_arg_or_env('--vault-passwords').split('||')
        except MissingInputException:
            vault_passwords = []

        num = 0
        opts = ''
        enforce_ask_pass = ctx.get_arg('--ask-vault-pass')

        for passwd in vault_passwords:
            num = num + 1

            if not passwd:
                continue

            if passwd.startswith('./') or passwd.startswith('/'):
                if os.path.isfile(passwd):
                    opts += ' --vault-password-file="%s" ' % (chdir + passwd)
                else:
                    self.io().error('Vault password file "%s" does not exist, calling --ask-vault-pass' % passwd)
                    enforce_ask_pass = True
            else:
                tmp_vault_file = './.rkd/.tmp-vault-' + str(uuid4())

                with open(tmp_vault_file, 'w') as f:
                    f.write(passwd)

                opts += ' --vault-password-file="%s" ' % (chdir + tmp_vault_file)

        if enforce_ask_pass:
            opts += ' --ask-vault-pass '

        return opts

    def _clear_old_vault_temporary_files(self):
        self.sh('rm -f ./.rkd/.tmp-vault*', capture=True)

    @classmethod
    def _add_vault_arguments_to_argparse(cls, parser: ArgumentParser):
        parser.add_argument('--ask-vault-pass', '-v', help='Ask for vault password interactively')
        parser.add_argument('--vault-passwords', '-V', help='Vault passwords separated by "||" eg. 123||456')


class UpdateFilesTask(BaseDeploymentTask):
    """Updates an Ansible role and required configuration files.
    Warning: Overwrites existing files, but does not remove custom files in '.rkd/deployment' directory"""

    def get_name(self) -> str:
        return ':update'

    def get_group_name(self) -> str:
        return ':harbor:deployment:files'

    def format_task_name(self, name) -> str:
        return development_formatting(name)

    def get_declared_envs(self) -> Dict[str, str]:
        envs = super(BaseDeploymentTask, self).get_declared_envs()
        envs['VAULT_PASSWORDS'] = ''

        return envs

    def configure_argparse(self, parser: ArgumentParser):
        self._add_vault_arguments_to_argparse(parser)

    def run(self, context: ExecutionContext) -> bool:
        self._preserve_vault_parameters_for_usage_in_inner_tasks(context)

        try:
            return self.install_and_configure_role(force_update=True)

        except MissingDeploymentConfigurationError as e:
            self.io().error_msg(str(e))
            return False


class DeploymentTask(BaseDeploymentTask):
    """Deploys your project from GIT to a PRODUCTION server

All changes needs to be COMMITED and PUSHED to GIT server, the task does not copy local files.

The deployment task can be extended by environment variables and switches to make possible any customizations
such as custom playbook, custom role or a custom inventory. The environment variables from .env are considered.

Example usage:
    # deploy services matching profile "gateway", use password stored in .vault-apssword for Ansible Vault
    harbor :deployment:apply -V .vault-password --profile=gateway

    # another example with Vault, multiple passwords, and environment variable usage
    # (NOTICE: paths to password files must begin with "/" or "./")
    VAULT_PASSWORDS="./.vault-password-file||other-plain-text-password" harbor :deployment:apply

    # deploy from different branch
    harbor :deployment:apply --branch production_fix_1

    # use SSH-AGENT & key-based authentication by specifying path to private key
    harbor :deployment:apply --git-key=~/.ssh/id_rsa
    """

    def get_name(self) -> str:
        return ':apply'

    def get_group_name(self) -> str:
        return ':harbor:deployment'

    def format_task_name(self, name) -> str:
        return development_formatting(name)

    def get_declared_envs(self) -> Dict[str, str]:
        envs = super(DeploymentTask, self).get_declared_envs()
        envs['PLAYBOOK'] = 'harbor.playbook.yml'
        envs['INVENTORY'] = 'harbor.inventory.cfg'
        envs['GIT_KEY'] = ''
        envs['VAULT_PASSWORDS'] = ''

        return envs

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('--playbook', '-p', help='Playbook name', default='harbor.playbook.yml')
        parser.add_argument('--git-key', '-k', help='Path to private key for a git repository eg. ~/.ssh/id_rsa',
                            default='')
        parser.add_argument('--inventory', '-i', help='Inventory filename', default='harbor.inventory.cfg')
        parser.add_argument('--debug', '-d', action='store_true', help='Set increased logging for Ansible output')
        parser.add_argument('--branch', '-b', help='Git branch to deploy from', default='master')
        parser.add_argument('--profile', help='Harbor profile to filter out services that needs to be deployed',
                            default='')
        self._add_vault_arguments_to_argparse(parser)

    def run(self, context: ExecutionContext) -> bool:
        playbook_name = context.get_arg_or_env('--playbook')
        inventory_name = context.get_arg_or_env('--inventory')
        git_private_key_path = context.get_arg_or_env('--git-key')
        branch = context.get_arg('--branch')
        profile = context.get_arg('--profile')
        debug = context.get_arg('--debug')

        # keep the vault arguments for decryption of deployment.yml
        self._preserve_vault_parameters_for_usage_in_inner_tasks(context)

        if not self.role_is_installed_and_configured():
            self.io().error_msg('Deployment not configured. Use `harbor :deployment:role:update` first')
            return False

        try:
            self.install_and_configure_role(force_update=False)

        except MissingDeploymentConfigurationError as e:
            self.io().error_msg(str(e))
            return False

        pwd_backup = os.getcwd()
        pid = None

        try:
            command = ''
            opts = ''

            if git_private_key_path:
                sock, pid = self.spawn_ssh_agent()
                command += 'export SSH_AUTH_SOCK=%s; export SSH_AGENT_PID=%i; ssh-add %s; sleep 5; ' % \
                           (sock, pid, git_private_key_path)

            if debug:
                opts += ' -vv '

            opts += ' -e git_branch="%s" ' % branch
            opts += ' -e harbor_deployment_profile="%s" ' % profile
            opts += self._get_vault_opts(context, '../../')

            os.chdir(self.ansible_dir)
            command += 'ansible-playbook ./%s -i %s %s' % (
                playbook_name,
                inventory_name,
                opts
            )

            self.spawn_ansible(command)
        finally:
            os.chdir(pwd_backup)

            if pid:
                self.kill_ssh_agent(pid)

        return True

    def spawn_ansible(self, command):
        self.io().info('Spawning Ansible')
        return self.sh(command)

    def spawn_ssh_agent(self) -> Tuple[str, int]:
        out = subprocess.check_output('eval $(ssh-agent -s);echo "|${SSH_AUTH_SOCK}|${SSH_AGENT_PID}";', shell=True).decode('utf-8')
        parts = out.split('|')
        sock = parts[1]
        pid = int(parts[2].strip())

        self.io().debug('Spawned ssh-agent - sock=%s, pid=%i' % (sock, pid))

        return sock, pid

    def kill_ssh_agent(self, pid: int):
        self.io().debug('Clean up - killing ssh-agent at PID=%i' % pid)
        subprocess.check_call(['kill', str(pid)])


class CreateExampleDeploymentFileTask(HarborBaseTask):
    """Creates a example deployment.yml file"""

    def get_group_name(self) -> str:
        return ':harbor:deployment'

    def get_name(self) -> str:
        return ':create-example'

    def format_task_name(self, name) -> str:
        return development_formatting(name)

    def run(self, context: ExecutionContext) -> bool:
        if os.path.isfile('./deployment.yml') or os.path.isfile('./deployment.yaml'):
            self.io().error_msg('deployment.yml or deployment.yaml already exists')
            return False

        subprocess.check_call(['cp', HARBOR_ROOT + '/../examples/deployment.yml', './deployment.yml'])
        self.io().success_msg('File "deployment.yml" created.')
        self.io().print_line()
        self.io().info_msg('The example is initially adjusted to work with Vagrant test virtual machine.')
        self.io().info_msg(' - `harbor :deployment:vagrant -c "up --provision"` to bring machine up')
        self.io().info_msg(' - `harbor :deployment:apply --git-key=~/.ssh/id_rsa` to perform a test deployment')

        return True


class ManageVagrantTask(BaseDeploymentTask):
    """Controls a test virtual machine using Vagrant"""

    def get_group_name(self) -> str:
        return ':harbor:deployment'

    def get_name(self) -> str:
        return ':vagrant'

    def format_task_name(self, name) -> str:
        return development_formatting(name)

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('--cmd', '-c', required=True, help='Vagrant commandline')

    def run(self, context: ExecutionContext) -> bool:
        cmd = context.get_arg('--cmd')

        try:
            subprocess.check_call('cd %s && vagrant %s' % (self.ansible_dir, cmd), shell=True)

        except subprocess.CalledProcessError:
            return False

        return True


class EditVaultTask(BaseDeploymentTask):
    """Edits an encrypted file

Example usage:
    # edit ".env-prod" file
    harbor :vault:edit .env-prod --vault-passwords="./.vault-password"

    # usage of environment variable (NOTICE: paths to password files must begin with "/" or "./")
    VAULT_PASSWORDS="./.vault-password-file||second-some-plaintext-password-there" harbor :vault:edit .env-prod

    # use a different text editor (you can also put EDITOR variable to your .env file)
    EDITOR=vim harbor :vault:edit deployment.yml

HINT: You can avoid writing the path in commandline each time by putting `VAULT_PASSWORDS=./path-to-password-file.txt` to the .env file
HINT: You can store vault password file on encrypted flash drive, and make a symbolic link. Every time when you mount an encrypted drive you will gain access to the project
NOTICE: When at least one of vault password files does not exist, then there will be a password prompt
"""

    def get_group_name(self) -> str:
        return ':harbor:vault'

    def get_name(self) -> str:
        return ':edit'

    def get_declared_envs(self) -> Dict[str, str]:
        envs = super(BaseDeploymentTask, self).get_declared_envs()
        envs['VAULT_PASSWORDS'] = ''

        return envs

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('filename', help='Filename')
        self._add_vault_arguments_to_argparse(parser)

    def run(self, context: ExecutionContext) -> bool:
        vault_opts = self._get_vault_opts(context)
        filename = context.get_arg('filename')

        try:
            subprocess.check_call('ansible-vault edit %s %s' % (vault_opts, filename), shell=True)
        finally:
            self._clear_old_vault_temporary_files()

        return True


class EncryptVaultTask(BaseDeploymentTask):
    """Encrypts/Decrypts a file using strong AES-256 algorithm,
output files are suitable to be kept in GIT repository

See the documentation for :harbor:vault:edit task for general file encryption documentation
"""

    def get_group_name(self) -> str:
        return ':harbor:vault'

    def get_name(self) -> str:
        return ':encrypt'

    def get_declared_envs(self) -> Dict[str, str]:
        envs = super(BaseDeploymentTask, self).get_declared_envs()
        envs['VAULT_PASSWORDS'] = ''

        return envs

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('--decrypt', '-d', action='store_true', help='Decrypt instead of encrypting')
        parser.add_argument('filename', help='Filename')
        self._add_vault_arguments_to_argparse(parser)

    def run(self, context: ExecutionContext) -> bool:
        vault_opts = self._get_vault_opts(context)
        filename = context.get_arg('filename')
        mode = 'decrypt' if context.get_arg('--decrypt') else 'encrypt'

        try:
            self.sh('ansible-vault %s %s %s' % (mode, vault_opts, filename), capture=False)
        finally:
            self._clear_old_vault_temporary_files()

        return True


class EnvEncryptTask(BaseDeploymentTask):
    """Manages the encryption of .env-prod file

The .env-prod file is a file that could be kept securely in GIT repository while containing passwords
required for services to work.
"""

    def get_group_name(self) -> str:
        return ':harbor:env'

    def get_name(self) -> str:
        return ':encrypt'

    def format_task_name(self, name) -> str:
        return development_formatting(name)

    def get_declared_envs(self) -> Dict[str, str]:
        envs = super(BaseDeploymentTask, self).get_declared_envs()
        envs['VAULT_PASSWORDS'] = ''

        return envs

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('--decrypt', '-d', action='store_true', help='Decrypt instead of encrypting')
        self._add_vault_arguments_to_argparse(parser)

    def run(self, context: ExecutionContext) -> bool:
        vault_opts = self._get_vault_opts(context)
        mode = 'decrypt' if context.get_arg('--decrypt') else 'encrypt'

        src = '.env'
        dst = '.env-prod'

        if mode == 'decrypt':
            src = '.env-prod'
            dst = '.env'

        try:
            self.sh('cp %s %s-tmp' % (src, dst))
            self.sh('ansible-vault %s %s %s-tmp' % (mode, vault_opts, dst), capture=False)
            self.sh('mv %s-tmp %s' % (dst, dst))
        finally:
            self._clear_old_vault_temporary_files()

        if mode == 'encrypt':
            try:
                self.sh('git add %s' % dst)
            except:
                pass

        return True



import os
from rkd.syntax import TaskDeclaration
from rkd import main as rkd_main
from rkd.standardlib.env import GetEnvTask
from rkd.standardlib.env import SetEnvTask
from .tasks.diagnostic import ListContainersTask
from .tasks.running import StartTask
from .tasks.running import UpgradeTask
from .tasks.running import StopTask
from .tasks.running import StopAndRemoveTask
from .tasks.running import RestartTask
from .tasks.running import PullTask
from .tasks.service import LogsTask
from .tasks.service import AnalyzeServiceTask
from .tasks.service import InspectContainerTask
from .tasks.service import ExecTask
from .tasks.service import ServiceRemoveTask
from .tasks.service import ServiceUpTask
from .tasks.service import ServiceStopTask
from .tasks.service import WaitForServiceTask
from .tasks.listing import ListDefinedServices
from .tasks.deployment import DeploymentTask
from .tasks.deployment import UpdateFilesTask
from .tasks.deployment import CreateExampleDeploymentFileTask
from .tasks.deployment import ManageVagrantTask
from .tasks.deployment import EditVaultTask
from .tasks.deployment import EncryptVaultTask
from .tasks.deployment import EnvEncryptTask
from .tasks.diagnostic import DumpComposeArguments
from .tasks.diagnostic import DumpComposeConfigTask
from .tasks.gateway import ReloadGatewayTask
from .tasks.gateway import ShowSSLStatusTask
from .tasks.gateway import ForceReloadSSLTask
from .tasks.configsmanagement import ListConfigsTask
from .tasks.configsmanagement import EnableConfigTask
from .tasks.configsmanagement import DisableConfigTask
from .tasks.maintenance import MaintenanceOnTask
from .tasks.maintenance import MaintenanceOffTask
from .tasks.repositories import FetchRepositoryTask
from .tasks.repositories import SetPermissionsForWritableDirectoriesTask
from .tasks.repositories import ListRepositoriesTask
from .tasks.repositories import FetchAllRepositories
from .tasks.structure import CreateHarborStructureTask


def imports():
    return [
        TaskDeclaration(ListContainersTask()),
        TaskDeclaration(StartTask()),
        TaskDeclaration(UpgradeTask()),
        TaskDeclaration(StopTask()),
        TaskDeclaration(StopAndRemoveTask()),
        TaskDeclaration(ListDefinedServices()),
        TaskDeclaration(ServiceUpTask()),
        TaskDeclaration(ServiceStopTask()),
        TaskDeclaration(WaitForServiceTask()),
        TaskDeclaration(ServiceRemoveTask()),
        TaskDeclaration(ExecTask()),
        TaskDeclaration(LogsTask()),
        TaskDeclaration(AnalyzeServiceTask()),
        TaskDeclaration(InspectContainerTask()),
        TaskDeclaration(PullTask()),
        TaskDeclaration(RestartTask()),
        TaskDeclaration(ListConfigsTask()),
        TaskDeclaration(EnableConfigTask()),
        TaskDeclaration(DisableConfigTask()),
        TaskDeclaration(DumpComposeArguments()),
        TaskDeclaration(DumpComposeConfigTask()),

        # production-related
        TaskDeclaration(ReloadGatewayTask()),
        TaskDeclaration(ShowSSLStatusTask()),
        TaskDeclaration(ForceReloadSSLTask()),
        TaskDeclaration(MaintenanceOnTask()),
        TaskDeclaration(MaintenanceOffTask()),
        TaskDeclaration(DeploymentTask()),
        TaskDeclaration(UpdateFilesTask()),
        TaskDeclaration(CreateExampleDeploymentFileTask()),
        TaskDeclaration(ManageVagrantTask()),
        TaskDeclaration(EditVaultTask()),
        TaskDeclaration(EncryptVaultTask()),
        TaskDeclaration(EnvEncryptTask()),

        # git
        TaskDeclaration(FetchRepositoryTask()),
        TaskDeclaration(FetchAllRepositories()),
        TaskDeclaration(SetPermissionsForWritableDirectoriesTask()),
        TaskDeclaration(ListRepositoriesTask()),

        TaskDeclaration(CreateHarborStructureTask()),
        TaskDeclaration(GetEnvTask()),
        TaskDeclaration(SetEnvTask())
    ]


def env_or_default(env_name: str, default: str):
    return os.environ[env_name] if env_name in os.environ else default


def main():
    os.environ['RKD_PATH'] = os.path.dirname(os.path.realpath(__file__)) + '/internal:' + os.getenv('RKD_PATH', '')
    os.environ['RKD_WHITELIST_GROUPS'] = env_or_default('RKD_WHITELIST_GROUPS', ':env,:harbor,')
    os.environ['RKD_ALIAS_GROUPS'] = env_or_default('RKD_ALIAS_GROUPS', '->:harbor')
    os.environ['RKD_UI'] = env_or_default('RKD_UI', 'false')
    rkd_main()


if __name__ == '__main__':
    main()



from harbor.expressions import safe_eval
from harbor.test import BaseHarborTestClass


class SafeEvalTests(BaseHarborTestClass):
    """author: https://stackoverflow.com/a/48135793/6782994
    """

    def test_basic(self):
        self.assertEqual(safe_eval("1", {}), 1)

    def test_local(self):
        self.assertEqual(safe_eval("a", {'a': 2}), 2)

    def test_local_bool(self):
        self.assertEqual(safe_eval("a==2", {'a': 2}), True)

    def test_lambda(self):
        self.assertRaises(ValueError, safe_eval, "lambda : None", {'a': 2})

    def test_bad_name(self):
        self.assertRaises(ValueError, safe_eval, "a == None2", {'a': 2})

    def test_attr(self):
        self.assertRaises(AttributeError, safe_eval, "a.__dict__", {'a': 2})

    def test_eval(self):
        self.assertRaises(ValueError, safe_eval, "eval('os.exit()')", {})

    def test_exec(self):
        self.assertRaises(SyntaxError, safe_eval, "exec 'import os'", {})

    def test_multiply(self):
        self.assertRaises(ValueError, safe_eval, "'s' * 3", {})

    def test_power(self):
        self.assertRaises(ValueError, safe_eval, "3 ** 3", {})

    def test_comprehensions(self):
        self.assertRaises(ValueError, safe_eval, "[i for i in [1,2]]", {'i': 1})

    def test_pattern_to_lookup_labels(self):
        self.assertTrue(
            safe_eval(
                '"org.riotkit.useMaintenanceMode" in service["labels"] and service["labels"]["org.riotkit.useMaintenanceMode"]', {
                    "service": {
                        'labels': {
                            'org.riotkit.useMaintenanceMode': True
                        }
                    }
                }
            )
        )

    def test_python36_nameconstant(self):
        """In Python 3.8 the 'True' is classified as Constant, in Python 3.6 as a NameConstant"""

        self.assertTrue(safe_eval('True', {}))

    def test_pattern_using_basic_string_functions(self):
        self.assertTrue(safe_eval('"org.riotkit.replicas".startswith("org.riotkit")', {}))



import os
from harbor.test import BaseHarborTestClass
from harbor.tasks.deployment import CreateExampleDeploymentFileTask


class CreateExampleDeploymentFileTaskTest(BaseHarborTestClass):
    def test_functional_example_deployment_file_is_created_if_it_does_not_exists(self):
        """Verify that the structure is copied, and the files are rendered using JINJA2"""

        self.execute_task(CreateExampleDeploymentFileTask(), args={}, env={})
        self.assertTrue(os.path.isfile(self.get_test_env_subdirectory('') + '/deployment.yml'))

    def test_functional_example_deployment_cannot_be_created_twice(self):
        """Test that file created once cannot be overridden"""

        # prepare test data
        path = self.get_test_env_subdirectory('') + '/deployment.yml'
        with open(path, 'wb') as f:
            f.write('test: true'.encode('utf-8'))

        # execute action
        out = self.execute_task(CreateExampleDeploymentFileTask(), args={}, env={})

        # verify
        self.assertIn('deployment.yml or deployment.yaml already exists', out, msg='Expected error message')

        with open(path, 'rb') as f:
            self.assertIn('test: true', f.read().decode('utf-8'), msg='Original content should be there')




from rkd.exception import TaskException


class ProfileNotFoundException(TaskException):
    def __init__(self, path: str):
        super().__init__('"%s" profile not found' % path)


class ServiceNotFoundInYaml(TaskException):
    def __init__(self, name: str):
        super().__init__('Service "%s" not found in any of loaded docker-compose YAMLs' % name)


class ServiceNotFoundInYamlLookedByCriteria(TaskException):
    def __init__(self, criteria: str):
        super().__init__('No any service found matching those criteria: %s' % criteria)


class ServiceNotCreatedException(TaskException):
    def __init__(self, name: str, instance_num: int = None):
        super().__init__('Service "%s" (instance=%s) was not yet created by docker engine, please start it first' % (
            name, str(instance_num)
        ))


class ServiceNotRunningException(TaskException):
    def __init__(self, name: str, instance_num='last'):
        super().__init__('Service "%s" seems to be not running, at least at instance #%s' % (name, instance_num))


class ServiceNotReadyException(TaskException):
    def __init__(self, name: str, text: str, instance_num='last'):
        super().__init__('Service "%s" #%s is not ready yet. "%s" text not found in logs' % (
            name, instance_num, text
        ))


class DeploymentException(TaskException):
    pass


class MissingDeploymentConfigurationError(TaskException):
    def __init__(self):
        super().__init__('Deployment not configured - missing deployment.yml or deployment.yaml file')



import subprocess
from harbor.test import BaseHarborTestClass
from harbor.tasks.deployment import DeploymentTask
from harbor.tasks.deployment import UpdateFilesTask


class DeploymentTaskTest(BaseHarborTestClass):
    def test_functional_validates_if_structure_exists(self):
        """Verify that :deployment:apply requires :deployment:files:update to be called first
        """

        out = self.execute_task(DeploymentTask(),
                                args={'--playbook': 'harbor.playbook.yml',
                                      '--inventory': 'harbor.inventory.yml',
                                      '--git-key': '',
                                      '--branch': 'master',
                                      '--profile': '',
                                      '--debug': False,
                                      '--vault-passwords': '',
                                      '--ask-vault-pass': False},
                                env={})

        self.assertIn('Deployment not configured. Use `harbor :deployment:role:update` first', out)
        self.assertIn('TASK_EXIT_RESULT=False', out)

    def test_functional_passes_structure_validation_after_using_update_command(self):
        """Verify that when :deployment:files:update is called, then :deployment:apply, then the validation
        in :deployment:apply is passing - as the files are created by first task
        """

        # Prepare configuration files first
        self.prepare_valid_deployment_yml()
        update_task = UpdateFilesTask()
        update_task.download_roles = lambda *args, **kwargs: None
        self.execute_task(update_task,
                          args={
                              '--ask-vault-pass': False,
                              '--vault-passwords': ''
                          },
                          env={})

        ansible_call = []

        deployment_task = DeploymentTask()
        deployment_task.spawn_ansible = lambda *args, **kwargs: ansible_call.append(args)

        self.execute_task(deployment_task,
                          args={'--playbook': 'harbor.playbook.yml',
                                '--inventory': 'harbor.inventory.yml',
                                '--git-key': '',
                                '--branch': 'master',
                                '--profile': '',
                                '--debug': False,
                                '--vault-passwords': '',
                                '--ask-vault-pass': False},
                          env={})

        self.assertIn('ansible-playbook', ansible_call[0][0])
        self.assertIn('./harbor.playbook.yml', ansible_call[0][0])
        self.assertIn('-i harbor.inventory.yml', ansible_call[0][0])
        self.assertIn('-e git_branch="master"', ansible_call[0][0])
        self.assertIn('-e harbor_deployment_profile=""', ansible_call[0][0])

    def test_functional_decrypts_encrypted_deployment_yml_file_on_startup(self):
        """Check that deployment.yml file can be encrypted, and that given vault password in commandline will allow
        to decrypt the contents automatically"""

        # 0) Preparation of configuration files
        self.prepare_valid_deployment_yml()
        update_task = UpdateFilesTask()
        update_task.download_roles = lambda *args, **kwargs: None
        self.execute_task(update_task,
                          args={
                              '--ask-vault-pass': False,
                              '--vault-passwords': ''
                          },
                          env={})

        ansible_call = []
        passphrase_file_path = self.get_test_env_subdirectory('.rkd') + '/tmp-secret.txt'

        # 1) Write test password
        with open(passphrase_file_path, 'w') as f:
            f.write('International-Workers-Association')

        # 2) Encrypt a file
        deployment_yml_path = self.get_test_env_subdirectory('') + '/deployment.yml'
        subprocess.check_call(
            ['ansible-vault encrypt --vault-password-file=%s %s' % (passphrase_file_path, deployment_yml_path)],
            shell=True)

        # 3) Run a deployment while the deployment.yml is encrypted
        deployment_task = DeploymentTask()
        deployment_task.spawn_ansible = lambda *args, **kwargs: ansible_call.append(args)

        out = self.execute_task(deployment_task,
                                args={'--playbook': 'harbor.playbook.yml',
                                      '--inventory': 'harbor.inventory.yml',
                                      '--git-key': '',
                                      '--branch': 'master',
                                      '--profile': '',
                                      '--debug': False,
                                      '--vault-passwords': passphrase_file_path,
                                      '--ask-vault-pass': False},
                                env={})

        self.assertIn('--vault-password-file=', ansible_call[0][0])
        self.assertIn('.rkd/tmp-secret.txt', ansible_call[0][0])
        self.assertIn('TASK_EXIT_RESULT=True', out)





from harbor.test import BaseHarborTestClass
from harbor.service import ServiceDeclaration


class ServiceDeclarationTest(BaseHarborTestClass):
    def test_get_domains(self):
        """Test get_domains() with multiple domains"""

        example_service = {
            'image': 'nginx:1.19',
            'environment': {
                'VIRTUAL_HOST': 'abcf.net,www.abcf.net'
            }
        }

        service = ServiceDeclaration('anarchist_black_cross_federation', example_service)
        self.assertEqual(['abcf.net', 'www.abcf.net'], service.get_domains())

    def test_watchtower_getter_true(self):
        example_service = {
            'image': 'nginx:1.19',
            'labels': {
                'com.centurylinklabs.watchtower.enable': 'true'
            }
        }

        service = ServiceDeclaration('anarchist_black_cross_federation', example_service)
        self.assertTrue(service.is_using_watchtower())

    def test_watchtower_getter_default_value(self):
        example_service = {
            'image': 'nginx:1.19'
        }

        service = ServiceDeclaration('anarchist_black_cross_federation', example_service)
        self.assertFalse(service.is_using_watchtower())

    def test_watchtower_getter_false(self):
        example_service = {
            'image': 'nginx:1.19',
            'labels': {
                'com.centurylinklabs.watchtower.enable': 'false'
            }
        }

        service = ServiceDeclaration('anarchist_black_cross_federation', example_service)
        self.assertFalse(service.is_using_watchtower())

    def test_watchtower_getter_invalid_value_converts_to_false(self):
        example_service = {
            'image': 'nginx:1.19',
            'labels': {
                'com.centurylinklabs.watchtower.enable': 'INVALID VALUE!'
            }
        }

        service = ServiceDeclaration('anarchist_black_cross_federation', example_service)
        self.assertFalse(service.is_using_watchtower())

    def test_maintenance_mode_label_name(self):
        example_service = {
            'image': 'nginx:1.19',
            'labels': {
                'org.riotkit.useMaintenanceMode': 'True'
            }
        }

        service = ServiceDeclaration('anarchist_black_cross_federation', example_service)
        self.assertTrue(service.is_using_maintenance_mode())

    def test_get_ports(self):
        ports = [
            '80:80',
            '443:443',
            '8001:80001'
        ]

        example_service = {
            'image': 'nginx:1.19',
            'ports': ports
        }

        service = ServiceDeclaration('phillyabc', example_service)
        self.assertEqual(ports, service.get_ports())

    def test_get_desired_replicas_count_returns_one_as_default(self):
        example_service = {
            'image': 'nginx:1.19'
        }

        service = ServiceDeclaration('phillyabc', example_service)
        self.assertEqual(1, service.get_desired_replicas_count())

    def test_get_desired_replicas_count_returns_declared_count(self):
        example_service = {
            'image': 'nginx:1.19',
            'labels': {
                'org.riotkit.replicas': '5'
            }
        }

        service = ServiceDeclaration('phillyabc', example_service)
        self.assertEqual(5, service.get_desired_replicas_count())

    def test_get_update_strategy_returns_declared_default(self):
        example_service = {
            'image': 'nginx:1.19'
        }

        service = ServiceDeclaration('phillyabc', example_service)
        self.assertEqual('compose', service.get_update_strategy('compose'))

    def test_get_update_strategy_returns_declared_value_in_yaml(self):
        example_service = {
            'image': 'nginx:1.19',
            'labels': {
                'org.riotkit.updateStrategy': 'recreate'
            }
        }

        service = ServiceDeclaration('phillyabc', example_service)
        self.assertEqual('recreate', service.get_update_strategy('compose'))

    def test_get_image_returns_declared_image(self):
        example_service = {
            'image': 'nginx:1.19'
        }

        service = ServiceDeclaration('iwa_ait', example_service)
        self.assertEqual('nginx:1.19', service.get_image())

    def test_get_image_returns_local_build_string_when_no_image_specified(self):
        example_service = {}

        service = ServiceDeclaration('iwa_ait', example_service)
        self.assertEqual('_docker_build_local:latest', service.get_image())

    def test_get_declared_version_returns_build_version_when_no_image_declared(self):
        example_service = {}

        service = ServiceDeclaration('iwa_ait', example_service)
        self.assertEqual('latest (build)', service.get_declared_version())

    def test_get_declared_version_returns_latest_when_image_declared_without_version(self):
        example_service = {
            'image': 'quay.io/riotkit/wp-auto-update'
        }

        service = ServiceDeclaration('iwa_ait', example_service)
        self.assertEqual('latest', service.get_declared_version())




from harbor.cached_loader import CachedLoader
from harbor.test import BaseHarborTestClass


class CachedLoaderTest(BaseHarborTestClass):
    def test_cached_calls_loader_once(self):
        test = 0

        def loader(num: int):
            return num + 1

        # called triple, but number increased once - the result is cached so the callback is not called more times
        test = CachedLoader.cached('test', lambda: loader(test))
        test = CachedLoader.cached('test', lambda: loader(test))
        test = CachedLoader.cached('test', lambda: loader(test))

        self.assertEqual(test, 1)



"""
Operations on containers

Interacts directly with Docker and Docker-Compose. In the future it will be possible to integrate also Docker Swarm.

:author: Blackandred <riotkit .. riseup.net>
"""

import os
import re
import subprocess
from time import time
from contextlib import contextmanager
from typing import Optional
from typing import Dict
from typing import List
from collections import OrderedDict
from json import loads as json_loads
from rkd.contract import ExecutionContext
from .interface import HarborTaskInterface
from .service import ServiceDeclaration
from .exception import ServiceNotReadyException
from .exception import ServiceNotCreatedException


class InspectedContainer(object):
    """Running or stopped container model retrieved from docker inspection"""

    name: str
    inspection: dict

    def __init__(self, name: str, inspection: dict):
        self.name = name
        self.inspection = inspection

    def get_id(self) -> str:
        return self.inspection['Id']

    def get_health_status(self) -> str:
        try:
            return self.inspection['State']['Health']['Status']
        except KeyError:
            return 'healthy' if self.inspection['State']['Status'] == 'running' else 'unhealthy'

    def has_health_check(self):
        try:
            self.inspection['State']['Health']['Status']
        except KeyError:
            return False

        return True

    def get_health_check_command(self) -> Optional[str]:
        try:
            return ' '.join(self.inspection['Config']['Healthcheck']['Test'][1:])
        except KeyError:
            return None

    def get_name(self) -> str:
        return self.name

    def get_image(self) -> Optional[str]:
        try:
            return self.inspection['Config']['Image']
        except KeyError:
            return None

    def get_start_time(self) -> str:
        return self.inspection['State']['StartedAt'][0:19]

    def to_dict(self) -> dict:
        return self.inspection


def build_compose_files_list(src_root: str, is_dev: bool) -> list:
    """Lists all YAML files to include in docker-compose arguments
    """

    yamls = {
        'conf': ['docker-compose.yml'],
        'conf.dev': []
    }

    for env_type in yamls.keys():
        for root, subdirs, files in os.walk(src_root + '/' + env_type):
            for file in files:
                if not file.endswith('.yml') and not file.endswith('.yaml'):
                    continue

                yamls[env_type].append(root + '/' + file)

    if is_dev:
        return yamls['conf'] + yamls['conf.dev']

    return yamls['conf']


class ComposeDriver(object):
    """Performs container operations using docker-compose
    """

    scope: HarborTaskInterface
    ctx: ExecutionContext
    project_name: str

    # lazy
    _compose_args: str = None

    def __init__(self, scope: HarborTaskInterface, ctx: ExecutionContext, project_name: str):
        self.scope = scope
        self.project_name = project_name
        self.ctx = ctx

    def get_last_container_name_for_service(self, service_name: str) -> str:
        """Gets full container name of last deployed service (last instance name)
        """

        service_name = self.project_name + '_' + service_name + '_'

        try:
            ps = self.scope.sh('docker ps --format=\'{{ .Names }}\' | grep "%s"' % service_name, capture=True).split("\n")
        except subprocess.CalledProcessError:
            ps = ''

        instance_numbers = []

        for instance in ps:
            matches = re.findall(service_name + '([0-9]+)', instance)

            if matches:
                instance_numbers.append(int(matches[0]))

        if not instance_numbers:
            raise ServiceNotCreatedException(service_name)

        return service_name + str(max(instance_numbers))

    def inspect_container(self, container_name: str):
        """Inspects a running/stopped container"""

        out = self.scope.sh('docker inspect %s' % container_name, capture=True)
        as_json = json_loads(out)

        if not as_json:
            raise Exception('Cannot inspect container, unknown docker inspect output: %s' % out)

        return InspectedContainer(container_name, as_json[0])

    def inspect_containers(self, names: list):
        """Inspect multiple containers by name at once (does same as inspect_container()
        but has better performance for multiple containers at once)
        """

        out = self.scope.sh('docker inspect %s' % ' '.join(names), capture=True)
        as_json = json_loads(out)

        if not as_json:
            raise Exception('Cannot inspect container, unknown docker inspect output: %s' % out)

        containers = []
        num = 0

        for sub_json in as_json:
            containers.append(InspectedContainer(names[num], sub_json))
            num += 1

        return containers

    @contextmanager
    def service_discovery_stopped(self):
        """Stops a service discovery for a moment"""

        try:
            self.scope.io().info('Suspending service discovery')
            self.compose(['stop', 'gateway_proxy_gen'])
            yield
        finally:
            self.scope.io().info('Starting service discovery')
            self.compose(['up', '-d', '--no-recreate', 'gateway_proxy_gen'])

    #
    # Methods to spawn processes in shell
    #
    def compose(self, arguments: list, capture: bool = False) -> Optional[str]:
        """Makes a call to docker-compose with all prepared arguments that should be"""

        cmd = 'IS_DEBUG_ENVIRONMENT=%s docker-compose %s %s' % (
            self.scope.is_dev_env,
            self.get_compose_args(),
            ' '.join(arguments)
        )
        self.scope.io().debug('Calling compose: %s' % cmd)

        return self.scope.sh(cmd, capture=capture)

    def exec_in_container(self, service_name: str, command: str, instance_num: int = None, capture: bool = True) -> str:
        """Executes a command in given container"""

        if instance_num is None:
            instance_num = int(self.get_last_container_name_for_service(service_name).split('_')[-1])

        return self.compose([
            'exec', '-T',
            '--index=%i' % instance_num if instance_num else '',
            service_name,
            'sh', '-c', '"', command.replace('"', '\\"'), '"'
        ], capture=capture)

    def exec_in_container_passthrough(self, command: str, service: ServiceDeclaration, instance_num: int = None,
                                      shell: str = '/bin/sh', tty: bool = True, interactive: bool = True):
        container_name = self.find_container_name(service, instance_num)

        opts = []

        if tty:
            opts.append('-t')

        if interactive:
            opts.append('-i')

        return subprocess.call(['docker', 'exec'] + opts + [container_name, shell, '-c', command])

    #
    # Basics - compose arguments present in all commands
    #
    def get_compose_args(self):
        """Gets arguments to use with docker-compose"""

        if not self._compose_args:
            self._compose_args = self.create_compose_arguments(self.ctx.get_env('APPS_PATH'),
                                                               is_dev=self.scope.is_dev_env)

            self.scope.io().debug('Compose args: %s' % self._compose_args)

        return self._compose_args

    def create_compose_arguments(self, src_root: str, is_dev: bool) -> str:
        """Internal method: Builds list of docker-compose arguments
        """

        yamls = build_compose_files_list(src_root, is_dev)
        args = ' --project-directory=%s -p %s ' % (os.getcwd(), self.project_name)

        for yaml_path in yamls:
            args += ' -f %s ' % yaml_path

        return args

    #
    # Domain specific methods
    #
    def up(self, service: ServiceDeclaration, norecreate: bool = False, force_recreate: bool = False,
           extra_args: str = '', capture: bool = False):
        """Bring up the service"""

        if norecreate and force_recreate:
            raise Exception('Logic exception, cannot set --no-recreate and --force-recreate at one time')

        self.compose([
            'up', '-d',
            '--no-recreate' if norecreate else '',
            '--force-recreate' if force_recreate else '',
            '--scale %s=%i' % (service.get_name(), service.get_desired_replicas_count()),
            service.get_name(),
            extra_args
        ], capture=capture)

    def find_container_name(self, service: ServiceDeclaration, instance_num: int = None) -> str:
        if not instance_num:
            return self.get_last_container_name_for_service(service.get_name())

        containers = self.get_created_containers(only_running=False)

        if not service.get_name() in containers:
            raise ServiceNotCreatedException(service.get_name(), instance_num=instance_num)

        matches = dict(containers[service.get_name()]).keys()

        if instance_num not in matches:
            raise ServiceNotCreatedException(service.get_name(), instance_num=instance_num)

        return self.create_container_name(service, instance_num)

    def create_container_name(self, service: ServiceDeclaration, instance_num: int) -> str:
        return self.project_name + '_' + service.get_name() + '_' + str(instance_num)

    def get_logs(self, service: ServiceDeclaration, instance_num: int = None, raw: bool = False,
                 follow: bool = False) -> str:
        """Gets logs from given container

        Args:
            service: Service declaration
            instance_num: Replica number
            raw: Do not return result, pass it directly to stdout and to stderr
            follow: Follow the output

        Returns:
            Logs in text format
        """

        args = ''

        if follow:
            args += ' --follow '

        command = 'docker logs %s "%s" 2>&1' % (args, self.find_container_name(service, instance_num))

        if raw:
            subprocess.call(command, shell=True)
            return ''

        return self.scope.sh(command, capture=True)

    def wait_for_log_message(self, text: str, service: ServiceDeclaration, instance_num: int = None,
                             timeout: int = 300) -> bool:

        """Waits for a text to appear in docker log

        Args:
            text: Text to wait for
            service: Service declaration
            instance_num: Replica number
            timeout: Timeout in seconds

        Raises:
            ServiceNotReadyException
        """

        timeout_at = time() + timeout

        while time() < timeout_at:
            logs = self.get_logs(service, instance_num)

            if text in logs:
                return True

        raise ServiceNotReadyException(service.get_name(), text, instance_num)

    def rm(self, service: ServiceDeclaration, extra_args: str = '', capture: bool = False):
        self.compose(['rm', '--stop', '--force', service.get_name(), extra_args], capture=capture)

    def kill_older_replica_than(self, service: ServiceDeclaration, project_name: str,
                                existing_containers: Dict[int, bool],
                                already_killed: int = 0):
        """Kill first old replica on the list"""

        instance_index = (already_killed * -1) - 1
        previous_instance_num = list(existing_containers.items())[instance_index][0]
        service_full_name = project_name + '_' + service.get_name() + '_' + str(previous_instance_num)

        self.scope.io().debug('Instances: ' + str(list(existing_containers.items())))
        self.scope.io().debug('Previous instance selector: %i' % instance_index)

        self.scope.io().info('Replica "%s" was spawned, killing older instance' % service_full_name)
        self.scope.io().info('Killing replica num=%i' % previous_instance_num)
        self.scope.sh('docker rm -f "%s"' % service_full_name)

    def scale_one_up(self, service: ServiceDeclaration) -> Dict[int, bool]:
        """Scale up and return last instance name (docker container name)"""

        desired_replicas = service.get_desired_replicas_count()
        self.scope.io().info('Scaling up to %i' % (desired_replicas + 1))

        try:
            self.compose(
                ['up', '-d', '--no-deps',
                 '--scale %s=%i' % (service.get_name(), desired_replicas + 1), service.get_name(), '2>&1'],
                capture=True
            )
        except subprocess.CalledProcessError as e:
            self.scope.io().err(e.output.decode('utf-8'))
            raise e

        self.scope.io().info('Finding last instance name...')
        instances: Dict[int, bool] = self.get_created_containers(only_running=False)[service.get_name()]

        self.scope.io().info('OK, it is "%s"' % max(instances.keys()))

        return instances

    def scale_to_desired_state(self, service: ServiceDeclaration):
        """Scale to declared state - eg. in case of a failure"""

        self.compose(
            ['up', '-d', '--no-deps',
             '--scale %s=%i' % (service.get_name(), service.get_desired_replicas_count()),
             service.get_name(),
             '2>&1']
        )

    def rm_image(self, img_to_remove: str, capture: bool = False):
        self.scope.sh('docker rmi %s 2>&1' % img_to_remove, capture=capture)

    def restart(self, service_name: str, extra_args: str = ''):
        self.compose(['restart', service_name, extra_args])

    def stop(self, service_name: str, extra_args: str = '', capture: bool = False):
        self.compose(['stop', service_name, extra_args], capture=capture)

    def ps(self, params: list):
        self.compose(['ps'] + params)

    def pull(self, services_names: list):
        self.compose(['pull'] + services_names)

    def get_created_containers(self, only_running: bool) -> Dict[str, Dict[int, bool]]:
        """Gets all running services"""

        # @todo: Cover with a test

        instances = self.scope.sh('docker ps -a --format="{{ .Names }}|{{ .Status }}"', capture=True).strip().split("\n")
        counted = {}

        for instance in instances:
            try:
                name, status = instance.split('|')
            except ValueError:
                continue

            if not name.startswith(self.project_name + '_'):
                continue

            service_name = name[len(self.project_name + '_'):-2]
            is_up = status.upper().startswith('UP')
            service_num = name.split('_')[-1]

            if service_name not in counted:
                counted[service_name] = OrderedDict()

            if only_running and not is_up:
                continue

            counted[service_name][int(service_num)] = is_up

        counted_and_sorted = {}

        for service in counted:
            counted_and_sorted[service] = OrderedDict(sorted(counted[service].items()))

        return counted_and_sorted

    def find_all_container_names_for_service(self, service: ServiceDeclaration) -> List[str]:
        """Finds all created container names for given service name

        Args:
            service: Service declaration object
        """

        created = self.get_created_containers(only_running=False)

        if not service.get_name() in created:
            raise ServiceNotCreatedException(service.get_name())

        return list(
            map(
                lambda instance_num: self.create_container_name(service, instance_num),
                created[service.get_name()].keys()
            )
        )



# license: MIT (C) tardyp
import ast


def safe_eval(expr, variables):
    """
    Safely evaluate a a string containing a Python
    expression.  The string or node provided may only consist of the following
    Python literal structures: strings, numbers, tuples, lists, dicts, booleans,
    and None. safe operators are allowed (and, or, ==, !=, not, +, -, ^, %, in, is)
    """
    _safe_names = {'None': None, 'True': True, 'False': False}
    _safe_nodes = [
        'Add', 'And', 'BinOp', 'BitAnd', 'BitOr', 'BitXor', 'BoolOp',
        'Compare', 'Dict', 'Eq', 'Expr', 'Expression', 'For',
        'Gt', 'GtE', 'Is', 'In', 'IsNot', 'LShift', 'List',
        'Load', 'Lt', 'LtE', 'Mod', 'Name', 'Not', 'NotEq', 'NotIn',
        'Num', 'Or', 'RShift', 'Set', 'Slice', 'Str', 'Sub',
        'Tuple', 'UAdd', 'USub', 'UnaryOp', 'boolop', 'cmpop',
        'expr', 'expr_context', 'operator', 'slice', 'unaryop', 'Constant', 'NameConstant',
        'Call', 'Attribute', 'Subscript', 'Index']
    node = ast.parse(expr, mode='eval')
    for subnode in ast.walk(node):
        subnode_name = type(subnode).__name__
        if isinstance(subnode, ast.Name):
            if subnode.id not in _safe_names and subnode.id not in variables:
                raise ValueError("Unsafe expression {}. contains {}".format(expr, subnode.id))
        if subnode_name not in _safe_nodes:
            raise ValueError("Unsafe expression {}. contains {}".format(expr, subnode_name))

    return eval(expr, variables)





import os
import json
from traceback import format_exc
from subprocess import CalledProcessError
from typing import Optional
from argparse import ArgumentParser
from rkd.contract import ExecutionContext
from .base import HarborBaseTask
from ..formatting import prod_formatting


class BaseRepositoryTask(HarborBaseTask):
    def get_group_name(self) -> str:
        return ':harbor:git:apps'

    def format_task_name(self, name) -> str:
        return prod_formatting(name)

    def get_app_repository_path(self, app_name: str, context: ExecutionContext) -> Optional[str]:
        return self.get_apps_path(context) + '/repos-enabled/%s.sh' % app_name

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('name', help='Application name (based on data/repos-enabled/*.sh)')

    def contextual_sh(self, path: str, script: str, capture: bool = False):
        """Adds a context of pre/post hooks and configuration"""

        return self.sh(
            '''
                current_pwd=$(pwd);

                # define hooks to override
                pre_update() { return 0; }
                post_update() { return 0; }

                source ''' + path + ''';

                ''' + script + '''
            ''',
            capture=capture
        )

    @staticmethod
    def _get_var(project_vars: dict, path: str, var_name: str):
        if var_name not in project_vars:
            raise Exception('%s is not defined in %s, as well as in %s' % (var_name, path, '.env'))

        return project_vars[var_name]

    def _load_project_vars(self, path: str):
        return json.loads(
            self.contextual_sh(path, 'python3 -c "import json; import os; print(json.dumps(dict(os.environ)))"',
                               capture=True)
        )

    def list_repositories(self, context: ExecutionContext) -> list:
        path = self.get_apps_path(context) + '/repos-enabled'
        names = []

        for file in os.scandir(path):
            file: os.DirEntry

            if not file.name.endswith('.sh'):
                continue

            names.append(file.name[:-3])

        names.sort()

        return names

    def _get_permissions_command(self):
        return 'sudo -E -u %s ' % self.app_user


class FetchRepositoryTask(BaseRepositoryTask):
    """Fetch a git repository from the remote"""

    def get_name(self) -> str:
        return ':update'

    def run(self, context: ExecutionContext) -> bool:
        path = self.get_app_repository_path(context.get_arg('--name'), context)

        if not os.path.isfile(path):
            self.io().error_msg('Cannot pull a repository: Unknown application, "%s" file not found' % path)
            return False

        # load variables from shell into our current Python context
        project_vars = self._load_project_vars(path)

        project_root_path = self.get_apps_path(context) + '/www-data/' + self._get_var(
            project_vars, path, 'GIT_PROJECT_DIR')

        # 1) run pre_update hook
        self.contextual_sh(path, 'pre_update %s' % project_root_path)

        # 2a) clone fresh repository
        if not os.path.isdir(project_root_path):
            self.io().info('Cloning a fresh repository at "%s"' % project_root_path)

            self._clone_new_repository(path, project_vars)

        # 2b) update existing repository - pull changes
        else:
            self.io().info('Pulling in an existing repository at "%s"' % project_root_path)
            self._pull_changes_into_existing_repository(path, project_vars)

        # 3) run pre_update hook
        self.contextual_sh(path, 'post_update %s' % project_root_path)

        self.io().print_opt_line()
        self.io().success_msg('Application\'s repository updated.')

        return True

    def _pull_changes_into_existing_repository(self, config_path: str, project_vars):
        """Updates existing local repository with remote changes"""

        # pass variables from Python to early-validate instead of throwing bash undeclared-var errors
        git_proto = self._get_var(project_vars, config_path, 'GIT_PROTO')
        git_project_dir = self._get_var(project_vars, config_path, 'GIT_PROJECT_DIR')
        git_user = self._get_var(project_vars, config_path, 'GIT_USER')
        git_server = self._get_var(project_vars, config_path, 'GIT_SERVER')
        git_password = self._get_var(project_vars, config_path, 'GIT_PASSWORD')
        git_org_name = self._get_var(project_vars, config_path, 'GIT_ORG_NAME')
        git_project_name = self._get_var(project_vars, config_path, 'GIT_PROJECT_NAME')

        command = '''
                set -e
                cd "./apps/www-data/''' + git_project_dir + '''"
    
                echo " >> Setting remote origin"
                %sudo% git config pull.rebase false
                %sudo% git remote remove origin 2>/dev/null || true
                %sudo% git remote add origin ''' + git_proto + '''://''' + git_user + ''':''' + git_password + '''@''' + git_server + '''/''' + git_org_name + '''/''' + git_project_name + '''
                %sudo% git pull origin master
                %sudo% git remote remove origin 2>/dev/null || true

            '''

        return self.contextual_sh(
            config_path,
            command.replace('%sudo%', self._get_permissions_command())
        )

    def _clone_new_repository(self, config_path: str, project_vars: dict):
        """Clones a new repository"""

        # pass variables from Python to early-validate instead of throwing bash undeclared-var errors
        git_proto = self._get_var(project_vars, config_path, 'GIT_PROTO')
        git_project_dir = self._get_var(project_vars, config_path, 'GIT_PROJECT_DIR')
        git_user = self._get_var(project_vars, config_path, 'GIT_USER')
        git_server = self._get_var(project_vars, config_path, 'GIT_SERVER')
        git_password = self._get_var(project_vars, config_path, 'GIT_PASSWORD')
        git_org_name = self._get_var(project_vars, config_path, 'GIT_ORG_NAME')
        git_project_name = self._get_var(project_vars, config_path, 'GIT_PROJECT_NAME')

        full_project_dir = os.path.realpath('./apps/www-data') + '/' + git_project_dir

        command = '''
                %sudo% git clone ''' + git_proto + '''://''' + git_user + ''':''' + git_password + '''@''' + git_server + '''/''' + git_org_name + '''/''' + git_project_name + ''' \
                    ''' + full_project_dir + ''' 
            '''

        command = command.replace('%sudo%', self._get_permissions_command())

        return self.contextual_sh(
            config_path,
            command
        )


class SetPermissionsForWritableDirectoriesTask(BaseRepositoryTask):
    """Make sure that the application would be able to write to allowed directories (eg. upload directories)"""

    def get_name(self) -> str:
        return ':set-permissions'

    def run(self, context: ExecutionContext) -> bool:
        path = self.get_app_repository_path(context.get_arg('--name'), context)

        if not os.path.isfile(path):
            self.io().error_msg('Cannot pull a repository: Unknown application, "%s" file not found' % path)
            return False

        # load variables from shell into our current Python context
        project_vars = self._load_project_vars(path)

        writable_dirs = self._get_var(project_vars, path, 'WRITABLE_DIRS').replace('\\ ', '@SPACE@').split(' ')
        project_root_path = self.get_apps_path(context) + '/www-data/' + self._get_var(
            project_vars, path, 'GIT_PROJECT_DIR')
        container_user = self._get_var(project_vars, path, 'DEFAULT_CONTAINER_USER')

        for writable_dir in writable_dirs:
            writable_dir = writable_dir.replace('@SPACE@', ' ')

            command = 'sudo chown %s "%s/%s"' % (container_user, project_root_path, writable_dir)

            self.io().info(command)
            self.sh(command)

        return True


class ListRepositoriesTask(BaseRepositoryTask):
    """List GIT repositories"""

    def configure_argparse(self, parser: ArgumentParser):
        pass

    def get_name(self) -> str:
        return ':list'

    def run(self, context: ExecutionContext) -> bool:
        header = ['Configured GIT repository']
        body = []

        for name in self.list_repositories(context):
            body.append([name])

        self.io().outln(self.table(header=header, body=body))

        return True


class FetchAllRepositories(BaseRepositoryTask):
    """List GIT repositories

    Ignores intermediate errors. When at least one repository fails to update,
    then overall status would be a failure.
    """

    def configure_argparse(self, parser: ArgumentParser):
        pass

    def get_name(self) -> str:
        return ':update-all'

    def run(self, context: ExecutionContext) -> bool:
        self.io().info('Fetching all repositories...')

        result = True

        with self.hooks_executed(context, 'repositories-upgrade'):
            for name in self.list_repositories(context):
                self.io().info('Updating "%s"' % name)

                try:
                    self.rkd([':harbor:git:apps:update', '--name=%s' % name])

                except CalledProcessError:
                    self.io().err(format_exc())
                    self.io().error_msg('Failed updating "%s"' % name)
                    result = False

        return result



from harbor.test import BaseHarborTestClass
from harbor.service import ServiceLocator
from harbor.exception import ServiceNotFoundInYaml


class ServiceLocatorTest(BaseHarborTestClass):
    def test_get_by_name_finds_service_and_returns_as_object(self):
        locator = ServiceLocator({
            'web_iwa_ait': {},
            'web_priamaakcia': {}
        })

        service = locator.get_by_name('web_priamaakcia')
        self.assertEqual('web_priamaakcia', service.get_name())

    def test_get_by_name_does_not_find_service_when_its_name_is_not_properly_specified(self):
        locator = ServiceLocator({
            'web_iwa_ait': {},
            'web_priamaakcia': {}
        })

        self.assertRaises(ServiceNotFoundInYaml, lambda: locator.get_by_name('non_existing'))



from harbor.test import BaseHarborTestClass


class TestRedirectToWWWFeature(BaseHarborTestClass):
    """Functional tests to check NGINX template feature - redirection from WWW to non-WWW hostname

    It's about a label: `org.riotkit.redirectFromWWW: true`
    """

    def test_redirects_from_www_to_non_www(self):
        """Verifies that there is a redirection www.nginx-redirect-to-www.local -> nginx-redirect-to-www.local"""

        drv = self._get_prepared_compose_driver()
        self.prepare_service_discovery(drv)
        self.prepare_example_service('website_with_redirect_to_www', uses_service_discovery=True)

        content = self.fetch_page_content('www.nginx-redirect-to-www.local')

        self.assertIn('Location: http://nginx-redirect-to-www.local', content)
        self.assertIn('HTTP/1.1 302', content)



from harbor.test import BaseHarborTestClass
from harbor.tasks.service import AnalyzeServiceTask
from harbor.service import ServiceDeclaration
from harbor.exception import ServiceNotFoundInYaml


class AnalyzeServiceTaskTest(BaseHarborTestClass):
    def test_report_shows_declared_and_actual_state_for_simple_service(self):
        service = ServiceDeclaration('alpine_3', {})

        drv = self._get_prepared_compose_driver()
        drv.up(service, capture=True)

        # assert it will not raise an error
        out = self.execute_task(AnalyzeServiceTask(), args={'name': 'alpine_3'})

        self.assertIn('Declared image:    alpine:3.11', out)
        self.assertIn('Update strategy:   compose', out)
        self.assertIn('Replicas:          1 of 1', out)
        self.assertIn('alpine:3.11', out)
        self.assertIn('env_simple_alpine_3_1', out)

    def test_shows_service_not_found_when_service_name_is_invalid(self):
        self.assertRaises(
            ServiceNotFoundInYaml,
            lambda: self.execute_task(AnalyzeServiceTask(), args={'name': 'invalid_name'})
        )




import os
from argparse import ArgumentParser
from abc import abstractmethod
from tabulate import tabulate
from rkd.contract import ExecutionContext
from .base import HarborBaseTask
from ..formatting import development_formatting


class ListConfigsTask(HarborBaseTask):
    """Gets environment variable value"""

    def get_name(self) -> str:
        return ':list'

    def get_group_name(self) -> str:
        return ':harbor:config'

    def format_task_name(self, name) -> str:
        return development_formatting(name)

    def run(self, context: ExecutionContext) -> bool:
        src_root = self.get_apps_path(context)
        rows = []

        for env_type in ['conf', 'conf.dev']:
            for root, subdirs, files in os.walk(src_root + '/' + env_type):
                for file in files:
                    # disabled
                    if file.endswith('.yml.disabled') or file.endswith('.yaml.disabled'):
                        rows.append([env_type + '/' + file[0:-9], 'No'])
                        continue

                    if not file.endswith('.yml') and not file.endswith('.yaml'):
                        continue

                    # enabled
                    rows.append([env_type + '/' + file, 'Yes'])

        self.io().out(tabulate(rows, headers=['Config file', 'Enabled']))
        self.io().print_line()

        return True


class AbstractManageConfigTask(HarborBaseTask):
    git_commands_suffix: str = ''  # used in tests
    git_mv_command = 'git mv'

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('--name', '-n', required=True, help='Configuration file name')

    def get_group_name(self) -> str:
        return ':harbor:config'

    def format_task_name(self, name) -> str:
        return development_formatting(name)

    def run(self, context: ExecutionContext) -> bool:
        config_name = context.get_arg('--name').replace('..', '')
        src_path = self.get_apps_path(context) + '/' + config_name + self.get_from_suffix()
        dst_path = self.get_apps_path(context) + '/' + config_name + self.get_dest_suffix()

        if not os.path.isfile(src_path):
            self.io().error_msg('Cannot find file at path "%s", check if name is correct' % src_path)
            return False

        self.sh((self.git_mv_command + ' "%s" "%s"' + self.git_commands_suffix) % (src_path, dst_path))
        self.sh(('git add "%s"' + self.git_commands_suffix) % dst_path)
        self.rkd([':harbor:config:list'])

        return True

    @staticmethod
    @abstractmethod
    def get_from_suffix():
        pass

    @staticmethod
    @abstractmethod
    def get_dest_suffix():
        pass


class EnableConfigTask(AbstractManageConfigTask):
    """Enable a configuration file - YAML"""

    def get_name(self) -> str:
        return ':enable'

    @staticmethod
    def get_from_suffix():
        return '.disabled'

    @staticmethod
    def get_dest_suffix():
        return ''


class DisableConfigTask(AbstractManageConfigTask):
    """Disable a configuration file - YAML"""

    def get_name(self) -> str:
        return ':disable'

    @staticmethod
    def get_from_suffix():
        return ''

    @staticmethod
    def get_dest_suffix():
        return '.disabled'



import subprocess
from argparse import ArgumentParser
from typing import Dict
from rkd.contract import ExecutionContext
from .base import HarborBaseTask
from ..formatting import prod_formatting


class GatewayBaseTask(HarborBaseTask):
    def configure_argparse(self, parser: ArgumentParser):
        super().configure_argparse(parser)

    def get_declared_envs(self) -> Dict[str, str]:
        envs = super().get_declared_envs()
        envs.update({
            'DISABLE_SSL': 'False'
        })

        return envs

    def make_sure_ssl_service_is_up(self, ctx: ExecutionContext):
        service = self.services(ctx).get_by_name('gateway_letsencrypt')

        self.containers(ctx).up(self.services(ctx).get_by_name('gateway'), norecreate=True)
        self.containers(ctx).up(self.services(ctx).get_by_name('gateway_proxy_gen'), norecreate=True)
        self.containers(ctx).up(service, norecreate=True)
        self.containers(ctx).wait_for_log_message('Sleep for', service, timeout=60)

    def format_task_name(self, name) -> str:
        return prod_formatting(name)


class ReloadGatewayTask(GatewayBaseTask):
    """Reload gateway, regenerate missing SSL certificates"""

    def run(self, ctx: ExecutionContext) -> bool:
        self.io().h2('Validating NGINX configuration')
        self.containers(ctx).exec_in_container('gateway', 'nginx -t')

        self.io().h2('Reloading NGINX configuration')
        self.containers(ctx).exec_in_container('gateway', 'nginx -s reload')

        if ctx.get_env('DISABLE_SSL').lower() != 'true':
            self.io().h2('Reloading SSL configuration')
            self.make_sure_ssl_service_is_up(ctx)
            self.containers(ctx).exec_in_container('gateway_letsencrypt', '/app/signal_le_service')

        return True

    def get_name(self) -> str:
        return ':reload'

    def get_group_name(self) -> str:
        return ':harbor:gateway'


class ShowSSLStatusTask(GatewayBaseTask):
    """Show status of SSL certificates"""

    def run(self, ctx: ExecutionContext) -> bool:
        if ctx.get_env('DISABLE_SSL').lower() != 'true':
            self.make_sure_ssl_service_is_up(ctx)
            self.io().out(self.containers(ctx).exec_in_container('gateway_letsencrypt', '/app/cert_status'))

        return True

    def get_name(self) -> str:
        return ':status'

    def get_group_name(self) -> str:
        return ':harbor:gateway:ssl'


class ForceReloadSSLTask(GatewayBaseTask):
    """Regenerate all certificates with force"""

    def run(self, ctx: ExecutionContext) -> bool:
        if ctx.get_env('DISABLE_SSL').lower() != 'true':
            self.make_sure_ssl_service_is_up(ctx)

            try:
                self.io().out(self.containers(ctx).exec_in_container('gateway_letsencrypt', '/app/force_renew'))

            except subprocess.CalledProcessError as err:
                self.io().error_msg(str(err))
                self.io().error('Output: ' + str(err.output))
                return False

        self.io().info_msg('SSL is disabled, not regenerating anything')
        return True

    def get_name(self) -> str:
        return ':regenerate'

    def get_group_name(self) -> str:
        return ':harbor:gateway:ssl'



import os
import subprocess
from tempfile import NamedTemporaryFile
from harbor.test import BaseHarborTestClass
from harbor.tasks.gateway import ForceReloadSSLTask


class TestForceReloadSSLTask(BaseHarborTestClass):
    def test_functional_force_reload_is_triggered_when_ssl_is_not_disabled(self):
        """We can only check if the container is taking the query

        Checks:
            - Required containers will be bringed up
            - Script will be triggered
        """

        # 1) Replace a script inside container
        #    As we do not want to trigger real SSL renew script - it would take long and fail anyway as it requires
        #    external service to pass in impossible conditions
        with NamedTemporaryFile(delete=False) as f:
            f.write('''#!/bin/bash
            echo "Renewing certificates..."
            exit 0
            '''.encode('utf-8'))

        subprocess.call(['chmod', '+x', f.name])

        self.mock_compose({
            'services': {
                'gateway_letsencrypt': {
                    'volumes': [
                        f.name + ':/app/force_renew'
                    ]
                }
            }
        })

        try:
            # 2) Test
            out = self.execute_task(ForceReloadSSLTask(), args={}, env={'DISABLE_SSL': ''})
            self.assertIn('Renewing certificates...', out)
        finally:
            # debug: travis
            print('TRAVIS DEBUG')
            subprocess.check_call('docker logs env_simple_gateway_letsencrypt_1', shell=True)

            self.remove_all_containers()
            os.unlink(f.name)

    def test_functional_ssl_regeneration_is_not_triggered_when_ssl_is_disabled(self):
        """Checks that SSL container will not be touched, when DISABLE_SSL=true
        """

        out = self.execute_task(ForceReloadSSLTask(), args={}, env={'DISABLE_SSL': 'true'})
        self.assertIn('not regenerating anything', out)



import subprocess
import json
from time import time
from time import sleep
from contextlib import contextmanager
from argparse import ArgumentParser
from rkd.contract import ExecutionContext
from .base import HarborBaseTask
from .base import UpdateStrategy
from ..exception import ServiceNotCreatedException
from ..service import ServiceDeclaration


class BaseHarborServiceTask(HarborBaseTask):
    """Abstract class"""

    def get_group_name(self) -> str:
        return ':harbor:service'

    def configure_argparse(self, parser: ArgumentParser):
        super().configure_argparse(parser)

        parser.add_argument('name', help='Service name')
        parser.add_argument('--extra-args', '-c', help='Optional compose arguments', default='')

    def prepare_single_for_single_container(self, ctx: ExecutionContext) -> tuple:
        service_name = ctx.get_arg('--name')
        service = self.services(ctx).get_by_name(service_name)
        instance_num = int(ctx.get_arg('--instance-num')) if ctx.get_arg('--instance-num') else None
        container_name = self.containers(ctx).find_container_name(service, instance_num)

        if not container_name:
            self.io().error_msg('Container not found')
            return None, None, None

        return container_name, service, instance_num

    def get_all_images_for_service(self, ctx: ExecutionContext, service: ServiceDeclaration):
        try:
            container_names = self.containers(ctx).find_all_container_names_for_service(service)
            inspected = self.containers(ctx).inspect_containers(container_names)

            return list(map(lambda container: container.get_image(), inspected))

        except ServiceNotCreatedException:
            # case: the service is just created, wasn't started yet
            return []


class ServiceUpTask(BaseHarborServiceTask):
    """Starts a single service

    Strategies:
        1) rolling: Perform a rolling update - scale services into more instances, and turn off older
                    containers silently (without downtime, requires extra resources)

           NOTICE: Do not use with databases

        2) compose: Perform a standard docker-compose up-like, if service is up-to-date then does not update, does not remove volumes

        3) recreate: Removes container and creates a new one, does not affect volumes

        3) auto (default): Performs automatic selection basing on label "org.riotkit.updateStrategy", defaults to "compose"
    """

    def get_name(self) -> str:
        return ':up'

    def configure_argparse(self, parser: ArgumentParser):
        super().configure_argparse(parser)
        parser.add_argument('--dont-recreate', '-d', action='store_true', help='Don\'t recreate the container if' +
                                                                               ' already existing (compose only)')
        parser.add_argument('--strategy', '-s',
                            help='Enforce an update strategy (optional)',
                            default='auto',
                            type=UpdateStrategy, choices=list(UpdateStrategy))
        parser.add_argument('--remove-previous-images', action='store_true',
                            help='Remove previous images if service had changed docker image')

    def run(self, context: ExecutionContext) -> bool:
        service_name = context.get_arg('name')
        strategy = str(context.get_arg('--strategy'))
        remove_previous_images = bool(context.get_arg('--remove-previous-images'))
        service = self.services(context).get_by_name(service_name)

        strategies = {
            'rolling': lambda: self.deploy_rolling(service, context),
            'compose': lambda: self.deploy_compose_like(service, context),
            'recreate': lambda: self.deploy_recreate(service, context)
        }

        with self.hooks_executed(context, 'service-start-%s' % service_name):
            with self._old_images_clean_up(context, service, clear_images=remove_previous_images):
                if strategy in strategies:
                    return strategies[strategy]()
                else:
                    if strategy == 'auto':
                        strategy = service.get_update_strategy(default='compose')

                        if strategy in strategies:
                            return strategies[strategy]()

                    self.io().error_msg('Invalid strategy selected: %s' % strategy)
                    return False

    @contextmanager
    def _old_images_clean_up(self, ctx: ExecutionContext, service: ServiceDeclaration, clear_images: bool):
        """Collects images, performs callback, then removes old images
        """

        if not clear_images:
            yield
            return

        self.io().debug('Finding all container names for service "%s"' % service.get_name())
        images = self.get_all_images_for_service(ctx, service)

        self.io().debug('OK, images collected')
        yield

        for image in images:
            # docker build was run locally
            if image is None:
                continue

            try:
                self.io().info('Trying to clean up image "%s"' % image)
                self.containers(ctx).rm_image(image, capture=True)
            except:
                self.io().warn('Cannot clean up image: "%s" [ignored, may be in use]' % image)

    def deploy_rolling(self, service: ServiceDeclaration, ctx: ExecutionContext) -> bool:
        """Rolling-update (without a downtime)

        1. Stop service discovery (to not add our not-ready-yet container to the load balancing)
        2. Scale up service one by one
        3. When health check is OK, then turn off older instance
        4. Start service discovery
        5. Repeat until get replaced all of the replicas
        """

        self.io().info('Doing a "rolling" deployment for "%s"' % service.get_name())
        desired_replicas = service.get_desired_replicas_count()
        processed = 0

        for replica_num in range(1, desired_replicas + 1):
            self.io().info('Processing instance #%i/%i' % (replica_num, desired_replicas))

            with self.containers(ctx).service_discovery_stopped():
                try:
                    existing_containers = self.containers(ctx).scale_one_up(service)

                    self.rkd([
                        '--no-ui',
                        ':harbor:service:wait-for',
                        service.get_name(),
                        '--instance=%s' % max(existing_containers.keys())
                    ])

                    self.containers(ctx).kill_older_replica_than(
                        service,
                        self.get_project_name(ctx),
                        existing_containers,
                        processed + 1
                    )

                except Exception as e:
                    self.io().error('Scaling back to declared state as error happened: %s' % str(e))
                    self.containers(ctx).scale_to_desired_state(service)
                    raise e

                processed += 1

            self.io().print_opt_line()

        return True

    def deploy_compose_like(self, service: ServiceDeclaration, ctx: ExecutionContext) -> bool:
        """Regular docker-compose up deployment (with downtime)"""

        self.io().info('Performing "compose" deployment for "%s"' % service.get_name())
        self.containers(ctx).up(service,
                                norecreate=bool(ctx.get_arg('--dont-recreate')),
                                extra_args=ctx.get_arg('--extra-args'))

        return True

    def deploy_recreate(self, service: ServiceDeclaration, ctx: ExecutionContext) -> bool:
        """Regular docker-compose up deployment (with downtime)"""

        self.io().info('Performing "recreate" deployment for "%s"' % service.get_name())
        self.containers(ctx).up(service,
                                norecreate=False, force_recreate=True,
                                extra_args=ctx.get_arg('--extra-args'))

        return True


class ServiceRemoveTask(BaseHarborServiceTask):
    """Stops and removes a container and it's images

    Use --with-image to remove images of all service instances
    """

    def get_name(self) -> str:
        return ':rm'

    def configure_argparse(self, parser: ArgumentParser):
        super().configure_argparse(parser)
        parser.add_argument('--with-image', '-w', help='Remove also images', action='store_true')

    def run(self, ctx: ExecutionContext) -> bool:
        service_name = ctx.get_arg('name')
        service = self.services(ctx).get_by_name(service_name)
        with_image = ctx.get_arg('--with-image')
        images = self.get_all_images_for_service(ctx, service) if with_image else []

        self.containers(ctx).rm(service, extra_args=ctx.get_arg('--extra-args'))

        for image in images:
            try:
                self.containers(ctx).rm_image(image, capture=True)
            except:
                pass

        return True


class ServiceStopTask(BaseHarborServiceTask):
    """Brings down the service without deleting the container
    """

    def get_name(self) -> str:
        return ':stop'

    def run(self, ctx: ExecutionContext) -> bool:
        service_name = ctx.get_arg('name')
        self.containers(ctx).stop(service_name, extra_args=ctx.get_arg('--extra-args'))

        return True


class ExecTask(BaseHarborServiceTask):
    """Execute a command in a container"""

    def configure_argparse(self, parser: ArgumentParser):
        super().configure_argparse(parser)
        parser.add_argument('--instance-num', '-i', default=None, help='Instance number. If None, then will pick last.')
        parser.add_argument('--command', '-e', default='/bin/sh', help='Command to execute')
        parser.add_argument('--shell', '-s', default='/bin/sh', help='Shell to use eg. bash or sh')
        parser.add_argument('--no-tty', help='Do not allocate tty', action='store_false')
        parser.add_argument('--no-interactive', help='Do not open interactive session', action='store_false')

    def get_name(self) -> str:
        return ':exec'

    def run(self, ctx: ExecutionContext) -> bool:
        command = ctx.get_arg('--command')
        shell = ctx.get_arg('--shell')
        tty = bool(ctx.get_arg('--no-tty'))
        interactive = bool(ctx.get_arg('--no-interactive'))

        container_name, service, instance_num = self.prepare_single_for_single_container(ctx)

        if not service:
            return False

        if shell != '/bin/sh' and command == '/bin/sh':
            command = shell

        self.containers(ctx).exec_in_container_passthrough(
            command, service, instance_num, shell=shell, tty=tty, interactive=interactive
        )

        return True


class InspectContainerTask(BaseHarborServiceTask):
    """Inspect a single container"""

    def configure_argparse(self, parser: ArgumentParser):
        super().configure_argparse(parser)
        parser.add_argument('--instance-num', '-i', default=None, help='Instance number')

    def get_name(self) -> str:
        return ':inspect'

    def run(self, ctx: ExecutionContext) -> bool:
        container_name, service, instance_num = self.prepare_single_for_single_container(ctx)

        if not service:
            return False

        container = self.containers(ctx).inspect_container(container_name)

        self.io().outln(json.dumps(container.to_dict(), indent=4, sort_keys=True))
        return True


class AnalyzeServiceTask(BaseHarborServiceTask):
    """Report status of a service"""

    def get_name(self) -> str:
        return ':report'

    def run(self, ctx: ExecutionContext) -> bool:
        service_name = ctx.get_arg('name')
        service = self.services(ctx).get_by_name(service_name)

        if not service:
            self.io().error_msg('Service not found')
            return False

        try:
            containers = self.containers(ctx).find_all_container_names_for_service(service)
        except ServiceNotCreatedException:
            containers = []

        self.print_service_summary(ctx, service, len(containers))
        self.io().print_line()

        if containers:
            self.print_containers_summary(ctx, service, containers)

        return True

    def print_service_summary(self, ctx: ExecutionContext, service: ServiceDeclaration, replicas_active: int):
        summary_body = [
            ['Replicas:', '%i of %i' % (replicas_active, service.get_desired_replicas_count())],
            ['Update strategy:', service.get_update_strategy()],
            ['Declared image:', service.get_image()],
            ['Startup priority:', service.get_priority_number()]
        ]

        self.io().outln(self.table(
            [],
            summary_body
        ))

    def print_containers_summary(self, ctx: ExecutionContext, service: ServiceDeclaration, containers: list):
        inspected = self.containers(ctx).inspect_containers(containers)

        #
        # containers summary
        #
        summary_body = []

        for container in inspected:
            summary_body.append([
                container.get_name(),
                ('(!!) ' if service.get_image() != container.get_image() else '') + container.get_image(),
                container.get_health_status(),
                container.get_start_time()
            ])

        self.io().outln(self.table(
            ['Name', 'Actual image', 'Status', 'Started'],
            summary_body
        ))


class LogsTask(BaseHarborServiceTask):
    """Display logs"""

    def configure_argparse(self, parser: ArgumentParser):
        super().configure_argparse(parser)
        parser.add_argument('--instance-num', '-i', default=None, help='Instance number')
        parser.add_argument('--follow', '-f', help='Follow the output', action='store_true')
        parser.add_argument('--buffered', help='Buffer output', action='store_true')

    def get_name(self) -> str:
        return ':logs'

    def run(self, ctx: ExecutionContext) -> bool:
        container_name, service, instance_num = self.prepare_single_for_single_container(ctx)
        buffered = bool(ctx.get_arg('--buffered'))

        if not service:
            self.io().error_msg('Service not found')
            return False

        self.io().outln(
            self.containers(ctx).get_logs(service, instance_num, raw=not buffered, follow=bool(ctx.get_arg('--follow')))
        )

        return True


class WaitForServiceTask(BaseHarborServiceTask):
    """Wait for service to be online"""

    def get_name(self) -> str:
        return ':wait-for'

    def configure_argparse(self, parser: ArgumentParser):
        super().configure_argparse(parser)
        parser.add_argument('--instance', '-i', required=False, help='Instance name, full container name')
        parser.add_argument('--timeout', '-t', default='120', help='Timeout in seconds')

    def run(self, ctx: ExecutionContext) -> bool:
        service_name = ctx.get_arg('name')
        timeout = int(ctx.get_arg('--timeout'))
        instance_num = int(ctx.get_arg('--instance')) if ctx.get_arg('--instance') else None

        service = self.services(ctx).get_by_name(service_name)

        try:
            container_name = self.containers(ctx).find_container_name(service, instance_num)
            container = self.containers(ctx).inspect_container(container_name)
        except ServiceNotCreatedException as e:
            self.io().error_msg(str(e))
            return False

        started_at = time()

        self.io().info('Checking health of "%s" service - %s' % (service_name, container_name))

        if container.has_health_check():
            while True:
                if time() - started_at >= timeout:
                    self.io().error_msg('Timeout of %is reached.' % timeout)
                    return False

                health = container.get_health_status()

                # do not wait for result, do check manually in mean time
                if health == 'starting':
                    self.io().warn('Docker reports "starting" - performing a manual check, we wont wait for docker')

                    try:
                        command = container.get_health_check_command().replace('"', '\\"')

                        self.containers(ctx).exec_in_container(
                            service_name=service_name,
                            command='/bin/sh -c "%s"' % command,
                            instance_num=instance_num
                        )

                        self.io().success_msg('Service healthy after %is' % (time() - started_at))
                        return True

                    except subprocess.CalledProcessError as e:
                        self.io().debug(str(e.output)[0:128])
                        sleep(1)
                        continue

                if health == 'healthy' or health == 'running':
                    self.io().success_msg('Service healthy after %i' % (time() - started_at))
                    return True

                sleep(1)
        else:
            self.io().warn('Instance has no healthcheck defined!')

        return True
